""" Convolution and padding operators"""
from .base import *
import numpy as np
from .solver_utils import grb
from ..patches import unify_shape, compute_patches_stride_padding, is_shape_used, create_valid_mask
from .gradient_modules import Conv2dGrad

EPS = 1e-2

class BoundConv(Bound):
    def __init__(self, attr=None, inputs=None, output_index=0, options=None):
        super().__init__(attr, inputs, output_index, options)

        if len(attr['kernel_shape']) == 1:
            # for 1d conv
            assert (attr['pads'][0] == attr['pads'][1])
            self.padding = [attr['pads'][0]]
            self.F_conv = F.conv1d
            self.conv_dim = 1
        else:
            # for 2d conv
            assert (attr['pads'][0] == attr['pads'][2])
            assert (attr['pads'][1] == attr['pads'][3])
            self.padding = [attr['pads'][0], attr['pads'][1]]
            self.F_conv = F.conv2d
            self.conv_dim = 2

        self.stride = attr['strides']
        self.dilation = attr['dilations']
        self.groups = attr['group']
        if len(inputs) == 3:
            self.has_bias = True
        else:
            self.has_bias = False
        self.relu_followed = False
        self.patches_start = True
        self.mode = options.get("conv_mode", "matrix")
        # denote whether this Conv is followed by a ReLU
        # if self.relu_followed is False, we need to manually pad the conv patches.
        # If self.relu_followed is True, the patches are padded in the ReLU layer
        # and the manual padding is not needed.

    def forward(self, *x):
        # x[0]: input, x[1]: weight, x[2]: bias if self.has_bias
        bias = x[2] if self.has_bias else None

        output = self.F_conv(x[0], x[1], bias, self.stride, self.padding, self.dilation, self.groups)

        return output

    def bound_backward(self, last_lA, last_uA, *x, **kwargs):
        if self.is_input_perturbed(1):
            raise NotImplementedError(
                'Weight perturbation for convolution layers has not been implmented.')

        lA_y = uA_y = lA_bias = uA_bias = None
        weight = x[1].lower

        def _bound_oneside(last_A):
            if last_A is None:
                return None, 0
            if type(last_A) is OneHotC:
                # Conv layer does not support the OneHotC fast path. We have to create a dense matrix instead.
                last_A = onehotc_to_dense(last_A, dtype=weight.dtype)

            if type(last_A) == Tensor:
                shape = last_A.size()
                # when (W−F+2P)%S != 0, construct the output_padding
                if self.conv_dim == 2:
                    output_padding0 = (
                        int(self.input_shape[2]) - (int(self.output_shape[2]) - 1) * self.stride[0] + 2 *
                        self.padding[0] - 1 - (int(weight.size()[2] - 1) * self.dilation[0]))
                    output_padding1 = (
                        int(self.input_shape[3]) - (int(self.output_shape[3]) - 1) * self.stride[1] + 2 *
                        self.padding[1] - 1 - (int(weight.size()[3] - 1) * self.dilation[0]))
                    next_A = F.conv_transpose2d(
                        last_A.reshape(shape[0] * shape[1], *shape[2:]), weight, None,
                        stride=self.stride, padding=self.padding, dilation=self.dilation,
                        groups=self.groups, output_padding=(output_padding0, output_padding1))
                else:
                    # for 1d conv, we use conv_transpose1d()
                    output_padding = (
                            int(self.input_shape[2]) - (int(self.output_shape[2]) - 1) * self.stride[0] + 2 *
                            self.padding[0] - 1 - (int(weight.size()[2] - 1) * self.dilation[0]))
                    next_A = F.conv_transpose1d(
                        last_A.reshape(shape[0] * shape[1], *shape[2:]), weight, None,
                        stride=self.stride, padding=self.padding, dilation=self.dilation,
                        groups=self.groups, output_padding=output_padding)

                next_A = next_A.view(shape[0], shape[1], *next_A.shape[1:])
                if self.has_bias:
                    # sum_bias = (last_A.sum((3, 4)) * x[2].lower).sum(2)
                    sum_bias = torch.einsum('sbc...,c->sb', last_A, x[2].lower)
                else:
                    sum_bias = 0
                return next_A, sum_bias
            elif type(last_A) == Patches:
                # Here we build and propagate a Patch object with (patches, stride, padding)
                assert self.conv_dim == 2, 'Patches mode not supports conv1d so far.'
                assert type(last_A) == Patches
                if last_A.identity == 0:
                    # FIXME (09/20): Don't call it relu_followed. Instead, make this a property of A, called "padded" and propagate this property.
                    if not self.relu_followed:
                        # The last_A.patches was not padded, so we need to pad them here.
                        # If this Conv layer is followed by a ReLU layer, then the padding was already handled there and there is no need to pad again.
                        one_d_unfolded_r = create_valid_mask(self.output_shape, last_A.patches.device,
                                                             weight.dtype,
                                                             last_A.patches.shape[-2:],
                                                             last_A.stride,
                                                             last_A.inserted_zeros,
                                                             last_A.padding,
                                                             last_A.output_padding,
                                                             last_A.unstable_idx if last_A.unstable_idx else None)
                        patches = last_A.patches * one_d_unfolded_r
                    else:
                        patches = last_A.patches

                    if self.has_bias:
                        # bias is x[2] (lower and upper are the same), and has shape (c,).
                        # Patches either has [out_c, batch, out_h, out_w, c, h, w] or [unstable_size, batch, c, h, w].
                        sum_bias = torch.einsum('sb...chw,c->sb...', patches, x[2].lower)
                        # sum_bias has shape (out_c, batch, out_h, out_w) or (unstable_size, batch).
                    else:
                        sum_bias = 0

                    flattened_patches = patches.reshape(
                        -1, patches.size(-3), patches.size(-2), patches.size(-1))
                    pieces = F.conv_transpose2d(
                        flattened_patches, insert_zeros(weight, last_A.inserted_zeros)
                        , stride=self.stride)
                    # New patch size: (out_c, batch, out_h, out_w, c, h, w) or (unstable_size, batch, c, h, w).
                    pieces = pieces.view(
                        *patches.shape[:-3], pieces.size(-3), pieces.size(-2),
                        pieces.size(-1))

                elif last_A.identity == 1:
                    # New patches have size [out_c, batch, out_h, out_w, c, h, w] if it is not sparse.
                    # New patches have size [unstable_size, batch, c, h, w] if it is sparse.
                    if last_A.unstable_idx is not None:
                        pieces = weight.view(
                            weight.size(0), 1, weight.size(1), weight.size(2),
                            weight.size(3))
                        # Select based on the output channel (out_h and out_w are irrelevant here).
                        pieces = pieces[last_A.unstable_idx[0]]
                        # Expand the batch dimnension.
                        pieces = pieces.expand(-1, last_A.shape[1], -1, -1, -1)
                        # Do the same for the bias.
                        if self.has_bias:
                            sum_bias = x[2].lower[last_A.unstable_idx[0]].unsqueeze(-1)
                            # bias has shape (unstable_size, batch).
                            sum_bias = sum_bias.expand(-1, last_A.shape[1])
                        else:
                            sum_bias = 0
                    else:
                        assert weight.size(0) == last_A.shape[0]
                        pieces = weight.view(
                            weight.size(0), 1, 1, 1, weight.size(1), weight.size(2),
                            weight.size(3)).expand(-1, *last_A.shape[1:4], -1, -1, -1)
                        # The bias (x[2].lower) has shape (out_c,) need to make it (out_c, batch, out_h, out_w).
                        # Here we should transpose sum_bias to set the batch dim to 1, aiming to keep it consistent with the matrix version
                        if self.has_bias:
                            sum_bias = x[2].lower.view(-1, 1, 1, 1).expand(-1, *last_A.shape[1:4])
                        else:
                            sum_bias = 0
                else:
                    raise NotImplementedError()
                padding = last_A.padding if last_A is not None else (0, 0, 0, 0)  # (left, right, top, bottom)
                stride = last_A.stride if last_A is not None else (1, 1)
                inserted_zeros = last_A.inserted_zeros if last_A is not None else 0
                output_padding = last_A.output_padding if last_A is not None else (0, 0, 0, 0)

                padding, stride, output_padding = compute_patches_stride_padding(
                    self.input_shape, padding, stride, self.padding, self.stride,
                    inserted_zeros, output_padding)

                if (inserted_zeros == 0 and not is_shape_used(output_padding)
                    and pieces.shape[-1] > self.input_shape[-1]):  # the patches is too large and from now on, we will use matrix mode instead of patches mode.
                    # This is our desired matrix: the input will be flattend to (batch_size, input_channel*input_x * input_y) and multiplies on this matrix.
                    # After multiplication, the desired output is (batch_size, out_channel*output_x*output_y).
                    # A_matrix has size (batch, out_c*out_h*out_w, in_c*in_h*in_w)
                    A_matrix = patches_to_matrix(
                        pieces, self.input_shape[1:], stride, padding,
                        last_A.output_shape, last_A.unstable_idx)
                    # print(f'Converting patches to matrix: old shape {pieces.shape}, size {pieces.numel()}; new shape {A_matrix.shape}, size {A_matrix.numel()}')
                    if isinstance(sum_bias, Tensor) and last_A.unstable_idx is None:
                        sum_bias = sum_bias.transpose(0, 1)
                        sum_bias = sum_bias.reshape(sum_bias.size(0), -1).transpose(0,1)
                    A_matrix = A_matrix.transpose(0,1)  # Spec dimension at the front.
                    return A_matrix, sum_bias
                new_patches = last_A.create_similar(
                        pieces, stride=stride, padding=padding, output_padding=output_padding,
                        identity=0, input_shape=self.input_shape)
                # if last_A is last_lA:
                #     print(f'Conv : start_node {kwargs["start_node"].name} layer {self.name} {new_patches}')
                return new_patches, sum_bias
            else:
                raise NotImplementedError()

        lA_x, lbias = _bound_oneside(last_lA)
        uA_x, ubias = _bound_oneside(last_uA)
        return [(lA_x, uA_x), (lA_y, uA_y), (lA_bias, uA_bias)], lbias, ubias

    def build_solver(self, *v, model, C=None, model_type="mip", solver_pkg="gurobi"):
        if self.is_input_perturbed(1):
            raise NotImplementedError("Weight perturbation for convolution layers has not been implmented.")

        assert self.dilation == (1, 1) or self.dilation == [1, 1]
        # e.g., last layer input gurobi vars (3,32,32)
        gvars_array = np.array(v[0])
        # pre_layer_shape (1,3,32,32)
        pre_layer_shape = np.expand_dims(gvars_array, axis=0).shape
        # this layer shape (1,8,16,16)
        this_layer_shape = self.output_shape
        out_lbs, out_ubs = None, None
        if hasattr(self, "lower"):
            # self.lower shape (1,8,16,16)
            out_lbs = self.lower.detach().cpu().numpy()
            out_ubs = self.upper.detach().cpu().numpy()

        # current layer weight (8,3,4,4)
        this_layer_weight = v[1].detach().cpu().numpy()
        # current layer bias (8,)
        this_layer_bias = None
        if self.has_bias:
            this_layer_bias = v[2].detach().cpu().numpy()
        weight_shape2, weight_shape3 = this_layer_weight.shape[2], this_layer_weight.shape[3]
        padding0, padding1 = self.padding[0], self.padding[1]
        stride0, stride1 = self.stride[0], self.stride[1]

        new_layer_gurobi_vars = []
        new_layer_gurobi_constrs = []

        # precompute row and column index mappings

        # compute row mapping: from current row to input rows
        # vectorization of following code:
        # for out_row_idx in range(this_layer_shape[2]):
        #     ker_row_min, ker_row_max = 0, weight_shape2
        #     in_row_idx_min = -padding0 + stride0 * out_row_idx
        #     in_row_idx_max = in_row_idx_min + weight_shape2 - 1
        #     if in_row_idx_min < 0:
        #         ker_row_min = -in_row_idx_min
        #     if in_row_idx_max >= pre_layer_shape[2]:
        #         ker_row_max = ker_row_max - in_row_idx_max + pre_layer_shape[2] - 1
        #     in_row_idx_min, in_row_idx_max = max(in_row_idx_min, 0), min(in_row_idx_max,
        #                                                                  pre_layer_shape[2] - 1)
        in_row_idx_mins = np.arange(this_layer_shape[2]) * stride0 - padding0
        in_row_idx_maxs = in_row_idx_mins + weight_shape2 - 1
        ker_row_mins = np.zeros(this_layer_shape[2], dtype=int)
        ker_row_maxs = np.ones(this_layer_shape[2], dtype=int) * weight_shape2
        ker_row_mins[in_row_idx_mins < 0] = -in_row_idx_mins[in_row_idx_mins < 0]
        ker_row_maxs[in_row_idx_maxs >= pre_layer_shape[2]] = \
            ker_row_maxs[in_row_idx_maxs >= pre_layer_shape[2]] - in_row_idx_maxs[in_row_idx_maxs >= pre_layer_shape[2]]\
            + pre_layer_shape[2] - 1
        in_row_idx_mins = np.maximum(in_row_idx_mins, 0)
        in_row_idx_maxs = np.minimum(in_row_idx_maxs, pre_layer_shape[2] - 1)

        # compute column mapping: from current column to input columns
        # vectorization of following code:
        # for out_col_idx in range(this_layer_shape[3]):
        #     ker_col_min, ker_col_max = 0, weight_shape3
        #     in_col_idx_min = -padding1 + stride1 * out_col_idx
        #     in_col_idx_max = in_col_idx_min + weight_shape3 - 1
        #     if in_col_idx_min < 0:
        #         ker_col_min = -in_col_idx_min
        #     if in_col_idx_max >= pre_layer_shape[3]:
        #         ker_col_max = ker_col_max - in_col_idx_max + pre_layer_shape[3] - 1
        #     in_col_idx_min, in_col_idx_max = max(in_col_idx_min, 0), min(in_col_idx_max,
        #                                                                  pre_layer_shape[3] - 1)
        in_col_idx_mins = np.arange(this_layer_shape[3]) * stride1 - padding1
        in_col_idx_maxs = in_col_idx_mins + weight_shape3 - 1
        ker_col_mins = np.zeros(this_layer_shape[3], dtype=int)
        ker_col_maxs = np.ones(this_layer_shape[3], dtype=int) * weight_shape3
        ker_col_mins[in_col_idx_mins < 0] = -in_col_idx_mins[in_col_idx_mins < 0]
        ker_col_maxs[in_col_idx_maxs >= pre_layer_shape[3]] = \
            ker_col_maxs[in_col_idx_maxs >= pre_layer_shape[3]] - in_col_idx_maxs[in_col_idx_maxs >= pre_layer_shape[3]]\
            + pre_layer_shape[3] - 1
        in_col_idx_mins = np.maximum(in_col_idx_mins, 0)
        in_col_idx_maxs = np.minimum(in_col_idx_maxs, pre_layer_shape[3] - 1)

        neuron_idx = 0
        for out_chan_idx in range(this_layer_shape[1]):
            out_chan_vars = []
            for out_row_idx in range(this_layer_shape[2]):
                out_row_vars = []

                # get row index range from precomputed arrays
                ker_row_min, ker_row_max = ker_row_mins[out_row_idx], ker_row_maxs[out_row_idx]
                in_row_idx_min, in_row_idx_max = in_row_idx_mins[out_row_idx], in_row_idx_maxs[out_row_idx]

                for out_col_idx in range(this_layer_shape[3]):

                    # get col index range from precomputed arrays
                    ker_col_min, ker_col_max = ker_col_mins[out_col_idx], ker_col_maxs[out_col_idx]
                    in_col_idx_min, in_col_idx_max = in_col_idx_mins[out_col_idx], in_col_idx_maxs[out_col_idx]

                    # init linear expression
                    lin_expr = this_layer_bias[out_chan_idx] if self.has_bias else 0

                    # init linear constraint LHS implied by the conv operation
                    for in_chan_idx in range(this_layer_weight.shape[1]):

                        coeffs = this_layer_weight[out_chan_idx, in_chan_idx, ker_row_min:ker_row_max, ker_col_min:ker_col_max].reshape(-1)
                        gvars = gvars_array[in_chan_idx, in_row_idx_min:in_row_idx_max+1, in_col_idx_min:in_col_idx_max+1].reshape(-1)
                        if solver_pkg == 'gurobi':
                            lin_expr += grb.LinExpr(coeffs, gvars)
                        else:
                            for i in range(len(coeffs)):
                                try:
                                    lin_expr += coeffs[i] * gvars[i]
                                except TypeError:
                                    lin_expr += coeffs[i] * gvars[i].var

                    # init potential lb and ub, which helps solver to finish faster
                    out_lb = out_lbs[0, out_chan_idx, out_row_idx, out_col_idx] if out_lbs is not None else -float('inf')
                    out_ub = out_ubs[0, out_chan_idx, out_row_idx, out_col_idx] if out_ubs is not None else float('inf')
                    if out_ub - out_lb < EPS:
                        """
                            If the inferred lb and ub are too close, it could lead to floating point disagreement
                            between solver's inferred lb and ub constraints and the computed ones from ab-crown.
                            Such disagreement can lead to "infeasible" result from the solver for feasible problem.
                            To avoid so, we relax the box constraints.
                            This should not affect the solver's result correctness,
                            since the tighter lb and ub can be inferred by the solver.
                        """
                        out_lb, out_ub = (out_lb + out_ub - EPS) / 2., (out_lb + out_ub + EPS) / 2.

                    # add the output var and constraint
                    var = model.addVar(lb=out_lb, ub=out_ub,
                                            obj=0, vtype=grb.GRB.CONTINUOUS,
                                            # name=f'lay{layer_idx}_[{out_chan_idx}, {out_row_idx}, {out_col_idx}]')
                                            name=f'lay{self.name}_{neuron_idx}')
                    # model.addConstr(lin_expr == var, name=f'lay{layer_idx}_[{out_chan_idx}, {out_row_idx}, {out_col_idx}]_eq')
                    # new_layer_gurobi_constrs.append(
                        # model.addConstr(lin_expr == var, name=f'lay{self.name}_{neuron_idx}_eq'))
                    model.addConstr(lin_expr == var, name=f'lay{self.name}_{neuron_idx}_eq')
                    neuron_idx += 1

                    out_row_vars.append(var)
                out_chan_vars.append(out_row_vars)
            new_layer_gurobi_vars.append(out_chan_vars)

        self.solver_vars = new_layer_gurobi_vars
        # self.solver_constrs = new_layer_gurobi_constrs
        model.update()

    def interval_propagate(self, *v, C=None):
        if self.is_input_perturbed(1):
            raise NotImplementedError("Weight perturbation for convolution layers has not been implmented.")

        norm = Interval.get_perturbation(v[0])
        norm = norm[0]

        h_L, h_U = v[0]
        weight = v[1][0]
        bias = v[2][0] if self.has_bias else None

        if norm == torch.inf:
            mid = (h_U + h_L) / 2.0
            diff = (h_U - h_L) / 2.0
            weight_abs = weight.abs()
            deviation = self.F_conv(diff, weight_abs, None, self.stride, self.padding, self.dilation, self.groups)
        elif norm > 0:
            norm, eps = Interval.get_perturbation(v[0])
            # L2 norm, h_U and h_L are the same.
            mid = h_U
            # TODO: padding
            assert not isinstance(eps, torch.Tensor) or eps.numel() == 1
            deviation = torch.mul(weight, weight).sum((1, 2, 3)).sqrt() * eps
            deviation = deviation.unsqueeze(0).unsqueeze(-1).unsqueeze(-1)
        else: # Here we calculate the L0 norm IBP bound using the bound proposed in [Certified Defenses for Adversarial Patches, ICLR 2020]
            norm, eps, ratio = Interval.get_perturbation(v[0])
            mid = h_U
            k = int(eps)
            weight_sum = torch.sum(weight.abs(), 1)
            deviation = torch.sum(torch.topk(weight_sum.view(weight_sum.shape[0], -1), k)[0], dim=1) * ratio

            if self.has_bias:
                center = self.F_conv(mid, weight, v[2][0], self.stride, self.padding, self.dilation, self.groups)
            else:
                center = self.F_conv(mid, weight, None, self.stride, self.padding, self.dilation, self.groups)

            ss = center.shape
            deviation = deviation.repeat(ss[2] * ss[3]).view(-1, ss[1]).t().view(ss[1], ss[2], ss[3])

        center = self.F_conv(mid, weight, bias, self.stride, self.padding, self.dilation, self.groups)

        upper = center + deviation
        lower = center - deviation
        return lower, upper

    def bound_dynamic_forward(self, *x, max_dim=None, offset=0):
        if self.is_input_perturbed(1) or self.is_input_perturbed(2):
            raise NotImplementedError("Weight perturbation for convolution layers has not been implmented.")
        weight = x[1].lb
        bias = x[2].lb if self.has_bias else None
        x = x[0]
        w = x.lw
        b = x.lb
        shape = w.shape
        shape_wconv = [shape[0] * shape[1]] + list(shape[2:])
        def conv2d(input, weight, bias, stride, padding, dilation, groups):
            """ There may be some CUDA error (illegal memory access) when
            the batch size is too large. Thus split the input into several
            batches when needed. """
            max_batch_size = 50
            if input.device != torch.device('cpu') and input.shape[0] > max_batch_size:
                ret = []
                for i in range((input.shape[0] + max_batch_size - 1) // max_batch_size):
                    ret.append(self.F_conv(
                        input[i*max_batch_size:(i+1)*max_batch_size],
                        weight, bias, stride, padding, dilation, groups))
                return torch.cat(ret, dim=0)
            else:
                return self.F_conv(input, weight, bias, stride, padding, dilation, groups)
        w_new = conv2d(
            w.reshape(shape_wconv), weight, None, self.stride, self.padding,
            self.dilation, self.groups)
        w_new = w_new.reshape(shape[0], -1, *w_new.shape[1:])
        b_new = conv2d(
            b, weight, bias, self.stride, self.padding, self.dilation, self.groups)
        return LinearBound(w_new, b_new, w_new, b_new, x_L=x.x_L, x_U=x.x_U, tot_dim=x.tot_dim)

    def bound_forward(self, dim_in, *x):
        if self.is_input_perturbed(1) or self.is_input_perturbed(2):
            raise NotImplementedError("Weight perturbation for convolution layers has not been implmented.")

        weight = x[1].lb
        bias = x[2].lb if self.has_bias else None
        x = x[0]

        mid_w = (x.lw + x.uw) / 2
        mid_b = (x.lb + x.ub) / 2
        diff_w = (x.uw - x.lw) / 2
        diff_b = (x.ub - x.lb) / 2
        weight_abs = weight.abs()
        shape = mid_w.shape
        shape_wconv = [shape[0] * shape[1]] + list(shape[2:])
        deviation_w = self.F_conv(
            diff_w.reshape(shape_wconv), weight_abs, None,
            self.stride, self.padding, self.dilation, self.groups)
        deviation_b = self.F_conv(
            diff_b, weight_abs, None,
            self.stride, self.padding, self.dilation, self.groups)
        center_w = self.F_conv(
            mid_w.reshape(shape_wconv), weight, None,
            self.stride, self.padding, self.dilation, self.groups)
        center_b = self.F_conv(
            mid_b, weight, bias,
            self.stride, self.padding, self.dilation, self.groups)
        deviation_w = deviation_w.reshape(shape[0], -1, *deviation_w.shape[1:])
        center_w = center_w.reshape(shape[0], -1, *center_w.shape[1:])

        return LinearBound(
            lw = center_w - deviation_w,
            lb = center_b - deviation_b,
            uw = center_w + deviation_w,
            ub = center_b + deviation_b)

    def build_gradient_node(self, grad_upstream):
        node_grad = Conv2dGrad(
            self, self.inputs[1].param, self.stride, self.padding,
            self.dilation, self.groups)
        return node_grad, (grad_upstream,), []

    def update_requires_input_bounds(self):
        self._check_weight_perturbation()


class BoundConvTranspose(Bound):
    def __init__(self, attr=None, inputs=None, output_index=0, options=None):
        super().__init__(attr, inputs, output_index, options)
        assert (attr['pads'][0] == attr['pads'][2])
        assert (attr['pads'][1] == attr['pads'][3])

        self.stride = attr['strides']
        self.padding = [attr['pads'][0], attr['pads'][1]]
        self.dilation = attr['dilations']
        self.groups = attr['group']
        self.output_padding = [attr.get('output_padding', [0, 0])[0], attr.get('output_padding', [0, 0])[1]]
        assert len(attr['kernel_shape']) == 2  # 2d transposed convolution.
        if len(inputs) == 3:
            self.has_bias = True
        else:
            self.has_bias = False
        self.mode = options.get("conv_mode", "matrix")
        assert self.output_padding == [0, 0]
        assert self.dilation == [1, 1]
        assert self.stride[0] == self.stride[1]
        assert self.groups == 1

        self.F_convtranspose = F.conv_transpose2d

    def forward(self, *x):
        # x[0]: input, x[1]: weight, x[2]: bias if self.has_bias
        bias = x[2] if self.has_bias else None
        output = F.conv_transpose2d(x[0], x[1], bias, stride=self.stride, padding=self.padding, dilation=self.dilation, groups=self.groups, output_padding=self.output_padding)
        return output


    def bound_backward(self, last_lA, last_uA, *x, **kwargs):
        if self.is_input_perturbed(1):
            raise NotImplementedError("Weight perturbation for convolution layers has not been implmented.")

        lA_y = uA_y = lA_bias = uA_bias = None
        weight = x[1].lower
        assert weight.size(-1) == weight.size(-2)

        def _bound_oneside(last_A):
            if last_A is None:
                return None, 0
            if type(last_A) is OneHotC:
                # Conv layer does not support the OneHotC fast path. We have to create a dense matrix instead.
                last_A = onehotc_to_dense(last_A, dtype=weight.dtype)

            if type(last_A) == Tensor:
                shape = last_A.size()
                next_A = F.conv2d(last_A.reshape(shape[0] * shape[1], *shape[2:]), weight, None,
                                            stride=self.stride, padding=self.padding, dilation=self.dilation,
                                            groups=self.groups)
                next_A = next_A.view(shape[0], shape[1], *next_A.shape[1:])
                if self.has_bias:
                    sum_bias = (last_A.sum((3, 4)) * x[2].lower).sum(2)
                else:
                    sum_bias = 0
                return next_A, sum_bias
            elif type(last_A) == Patches:
                # Here we build and propagate a Patch object with (patches, stride, padding)
                assert type(last_A) == Patches
                if last_A.identity == 0:
                    patches = last_A.patches

                    # FIXME: so far, assume there will be a relu layer in its input.

                    if self.has_bias:
                        # bias is x[2] (lower and upper are the same), and has shape (c,).
                        # Patches either has [out_c, batch, out_h, out_w, c, h, w] or [unstable_size, batch, c, h, w].
                        sum_bias = torch.einsum('sb...chw,c->sb...', patches, x[2].lower)
                        # sum_bias has shape (out_c, batch, out_h, out_w) or (unstable_size, batch).
                    else:
                        sum_bias = 0

                    flattened_patches = patches.reshape(-1, patches.size(-3), patches.size(-2), patches.size(-1))
                    # Merge patches with this layer's weights. Weight must be flipped here; and if stride != 1, we must insert zeros in the input image.
                    # For conv_transpose2d, the weight matrix is in the (in, out, k, k) shape.
                    # pieces = F.conv_transpose2d(flattened_patches, weight.transpose(0,1).flip(-1,-2), stride=self.stride)
                    # pieces = F.conv_transpose2d(flattened_patches, weight.transpose(0,1).flip(-1,-2), stride=last_A.inserted_zeros + 1)
                    # Use padding in conv_transposed2d directly.
                    pieces = F.conv_transpose2d(
                            # Transpose because the weight has in_channel before out_channel.
                            flattened_patches, insert_zeros(weight.transpose(0,1).flip(-1,-2), last_A.inserted_zeros))
                    # New patch size: (out_c, batch, out_h, out_w, c, h, w) or (unstable_size, batch, c, h, w).
                    pieces = pieces.view(*patches.shape[:-3], pieces.size(-3), pieces.size(-2), pieces.size(-1))

                elif last_A.identity == 1:
                    # New patches have size [out_c, batch, out_h, out_w, c, h, w] if it is not sparse.
                    # New patches have size [unstable_size, batch, c, h, w] if it is sparse.
                    if last_A.unstable_idx is not None:
                        raise NotImplementedError()
                    else:
                        assert weight.size(0) == last_A.shape[0]
                        pieces = weight.view(weight.size(0), 1, 1, 1, weight.size(1), weight.size(2), weight.size(3)).expand(-1, *last_A.shape[1:4], -1, -1, -1)
                        # The bias (x[2].lower) has shape (out_c,) need to make it (out_c, batch, out_h, out_w).
                        # Here we should transpose sum_bias to set the batch dim to 1, aiming to keep it consistent with the matrix version
                        sum_bias = x[2].lower.view(-1, 1, 1, 1).expand(-1, *last_A.shape[1:4])
                else:
                    raise NotImplementedError()
                patches_padding = last_A.padding if last_A is not None else (0, 0, 0, 0)  # (left, right, top, bottom)
                output_padding = last_A.output_padding if last_A is not None else (0, 0, 0, 0)  # (left, right, top, bottom)
                inserted_zeros = last_A.inserted_zeros
                assert self.stride[0] == self.stride[1]

                # Unify the shape to 4-tuple.
                output_padding = unify_shape(output_padding)
                patches_padding = unify_shape(patches_padding)
                this_stride = unify_shape(self.stride)
                this_padding = unify_shape(self.padding)

                # Compute new padding. Due to the shape flip during merging, we need to check the string/size on the dimension 3 - j.
                # TODO: testing for asymmetric shapes.
                padding = tuple(p * (inserted_zeros + 1) + (weight.size(3 - j//2) - 1) for j, p in enumerate(patches_padding))

                # Compute new output padding
                output_padding = tuple(p * (inserted_zeros + 1) + this_padding[j] for j, p in enumerate(output_padding))
                # When we run insert_zeros, it's missing the right most column and the bottom row.
                # padding = (padding[0], padding[1] + inserted_zeros, padding[2], padding[3] + inserted_zeros)

                # If no transposed conv so far, inserted_zero is 0.
                # We a transposed conv is encountered, stride is multiplied on it.
                inserted_zeros = (inserted_zeros + 1) * this_stride[0] - 1

                # FIXME: disabled patches_to_matrix because not all parameters are supported.
                if inserted_zeros == 0 and not is_shape_used(output_padding) and pieces.shape[-1] > self.input_shape[-1]:  # the patches is too large and from now on, we will use matrix mode instead of patches mode.
                    # This is our desired matrix: the input will be flattend to (batch_size, input_channel*input_x * input_y) and multiplies on this matrix.
                    # After multiplication, the desired output is (batch_size, out_channel*output_x*output_y).
                    # A_matrix has size (batch, out_c*out_h*out_w, in_c*in_h*in_w)
                    assert inserted_zeros == 0
                    A_matrix = patches_to_matrix(pieces, self.input_shape[1:], last_A.stride, padding, last_A.output_shape, last_A.unstable_idx)
                    if isinstance(sum_bias, Tensor) and last_A.unstable_idx is None:
                        sum_bias = sum_bias.transpose(0, 1)
                        sum_bias = sum_bias.reshape(sum_bias.size(0), -1).transpose(0,1)
                    A_matrix = A_matrix.transpose(0,1)  # Spec dimension at the front.
                    return A_matrix, sum_bias
                new_patches = last_A.create_similar(
                        pieces, padding=padding, inserted_zeros=inserted_zeros, output_padding=output_padding,
                        input_shape=self.input_shape)
                # if last_A is last_lA:
                #     print(f'ConvT input : start_node {kwargs["start_node"].name} layer {self.name} {last_lA}')
                #     print(f'ConvT layer : padding {self.padding} stride {self.stride} kernel {list(weight.shape[-2:])} input {list(self.input_shape)} output {list(self.output_shape)}')
                #     print(f'ConvT output: start_node {kwargs["start_node"].name} layer {self.name} {new_patches}')
                return new_patches, sum_bias
            else:
                raise NotImplementedError()

        lA_x, lbias = _bound_oneside(last_lA)
        uA_x, ubias = _bound_oneside(last_uA)
        return [(lA_x, uA_x), (lA_y, uA_y), (lA_bias, uA_bias)], lbias, ubias

    def interval_propagate(self, *v, C=None):
        if self.is_input_perturbed(1):
            raise NotImplementedError("Weight perturbation for convolution layers has not been implmented.")

        norm = Interval.get_perturbation(v[0])
        norm = norm[0]

        h_L, h_U = v[0]
        weight = v[1][0]
        bias = v[2][0] if self.has_bias else None

        if norm == torch.inf:
            mid = (h_U + h_L) / 2.0
            diff = (h_U - h_L) / 2.0
            weight_abs = weight.abs()
            deviation = F.conv_transpose2d(diff, weight_abs, None, stride=self.stride, padding=self.padding, dilation=self.dilation, groups=self.groups, output_padding=self.output_padding)
        elif norm > 0:
            raise NotImplementedError()
            norm, eps = Interval.get_perturbation(v[0])
            # L2 norm, h_U and h_L are the same.
            mid = h_U
            # TODO: padding
            deviation = torch.mul(weight, weight).sum((1, 2, 3)).sqrt() * eps
            deviation = deviation.unsqueeze(0).unsqueeze(-1).unsqueeze(-1)
        else: # Here we calculate the L0 norm IBP bound using the bound proposed in [Certified Defenses for Adversarial Patches, ICLR 2020]
            raise NotImplementedError()

        center = F.conv_transpose2d(mid, weight, bias, stride=self.stride, padding=self.padding, dilation=self.dilation, groups=self.groups, output_padding=self.output_padding)

        upper = center + deviation
        lower = center - deviation
        return lower, upper

    def bound_forward(self, dim_in, *x):
        if self.is_input_perturbed(1) or self.is_input_perturbed(2):
            raise NotImplementedError("Weight perturbation for convolution layers has not been implmented.")

        weight = x[1].lb
        bias = x[2].lb if self.has_bias else None
        x = x[0]

        mid_w = (x.lw + x.uw) / 2
        mid_b = (x.lb + x.ub) / 2
        diff_w = (x.uw - x.lw) / 2
        diff_b = (x.ub - x.lb) / 2
        weight_abs = weight.abs()
        shape = mid_w.shape
        shape_wconv = [shape[0] * shape[1]] + list(shape[2:])
        deviation_w = self.F_convtranspose(
            diff_w.reshape(shape_wconv), weight_abs, None, output_padding=self.output_padding,
            stride=self.stride, padding=self.padding, dilation=self.dilation, groups=self.groups)
        deviation_b = self.F_convtranspose(
            diff_b, weight_abs, None, output_padding=self.output_padding,
            stride=self.stride, padding=self.padding, dilation=self.dilation, groups=self.groups)
        center_w = self.F_convtranspose(
            mid_w.reshape(shape_wconv), weight, output_padding=self.output_padding,
            stride=self.stride, padding=self.padding, dilation=self.dilation, groups=self.groups)
        center_b = self.F_convtranspose(
            mid_b, weight, bias, output_padding=self.output_padding,
            stride=self.stride, padding=self.padding, dilation=self.dilation, groups=self.groups)
        deviation_w = deviation_w.reshape(shape[0], -1, *deviation_w.shape[1:])
        center_w = center_w.reshape(shape[0], -1, *center_w.shape[1:])

        return LinearBound(
            lw = center_w - deviation_w,
            lb = center_b - deviation_b,
            uw = center_w + deviation_w,
            ub = center_b + deviation_b)

class BoundPad(Bound):
    def __init__(self, attr=None, inputs=None, output_index=0, options=None):
        super().__init__(attr, inputs, output_index, options)
        if hasattr(attr, 'pads'):
            self.padding = attr['pads'][2:4] + attr['pads'][6:8]
        else:
            self.padding = [0, 0, 0, 0]
        self.value = attr.get('value', 0.0)
        assert self.padding == [0, 0, 0, 0]

    def forward(self, x, pad, value=0.0):
        # TODO: padding for 3-D or more dimensional inputs.
        assert x.ndim == 4
        # x[1] should be [0,0,pad_top,pad_left,0,0,pad_bottom,pad_right]
        assert pad[0] == pad[1] == pad[4] == pad[5] == 0
        pad = [int(pad[3]), int(pad[7]), int(pad[2]), int(pad[6])]
        final = F.pad(x, pad, value=value)
        self.padding, self.value = pad, value
        return final

    def interval_propagate(self, *v):
        l, u = zip(*v)
        return Interval.make_interval(self.forward(*l), self.forward(*u), v[0])

    def bound_backward(self, last_lA, last_uA, *x, **kwargs):
        # TODO: padding for 3-D or more dimensional inputs.
        left, right, top, bottom = self.padding
        def _bound_oneside(last_A):
            if last_A is None:
                return None
            assert type(last_A) is Patches or last_A.ndim == 5
            if type(last_A) is Patches:
                if isinstance(last_A.padding, tuple):
                    new_padding = (last_A.padding[0] + left, last_A.padding[1] + right, last_A.padding[2] + top, last_A.padding[3] + bottom)
                else:
                    new_padding = (last_A.padding + left, last_A.padding + right, last_A.padding + top, last_A.padding + bottom)
                return last_A.create_similar(padding=new_padding)
            else:
                shape = last_A.size()
                return last_A[:, :, :, top:(shape[3] - bottom), left:(shape[4] - right)]
        last_lA = _bound_oneside(last_lA)
        last_uA = _bound_oneside(last_uA)
        return [(last_lA, last_uA), (None, None), (None, None)], 0, 0

    def build_solver(self, *v, model, C=None, model_type="mip", solver_pkg="gurobi"):
        # e.g., last layer input gurobi vars (3,32,32)
        gvars_array = np.array(v[0])
        # pre_layer_shape (1,3,32,32)
        pre_layer_shape = np.expand_dims(gvars_array, axis=0).shape
        # this layer shape (1,3,35,35)
        this_layer_shape = self.output_shape
        # v1 = tensor([0, 0, 1, 1, 0, 0, 2, 2])
        # [0,0,pad_top,pad_left,0,0,pad_bottom,pad_right]
        # => [left, right, top, bottom]
        padding = [int(v[1][3]), int(v[1][7]), int(v[1][2]), int(v[1][6])]
        left, right, top, bottom = padding
        assert pre_layer_shape[2] + padding[0] + padding[1] == this_layer_shape[2]
        assert pre_layer_shape[3] + padding[2] + padding[3] == this_layer_shape[3]

        new_layer_gurobi_vars = []
        neuron_idx = 0
        for out_chan_idx in range(this_layer_shape[1]):
            out_chan_vars = []
            for out_row_idx in range(this_layer_shape[2]):
                out_row_vars = []
                row_pad = out_row_idx < left or out_row_idx >= this_layer_shape[2] - right
                for out_col_idx in range(this_layer_shape[3]):
                    col_pad = out_col_idx < top or out_col_idx >= this_layer_shape[3] - bottom
                    if row_pad or col_pad:
                        v = model.addVar(lb=0, ub=0,
                                    obj=0, vtype=grb.GRB.CONTINUOUS,
                                    name=f'pad{self.name}_{neuron_idx}')
                    else:
                        v = gvars_array[out_chan_idx, out_row_idx - left, out_col_idx - top]
                    # print(out_chan_idx, out_row_idx, out_col_idx, row_pad, col_pad, v.LB, v.UB)
                    neuron_idx += 1

                    out_row_vars.append(v)
                out_chan_vars.append(out_row_vars)
            new_layer_gurobi_vars.append(out_chan_vars)

        self.solver_vars = new_layer_gurobi_vars
        model.update()



"""Edit the computational graph in BoundedModule."""

from .bound_ops import Bound

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .bound_general import BoundedModule


# Make sure the nodes already have `name` and `input_name`
def add_nodes(self: 'BoundedModule', nodes):
    # TODO check duplicate names
    nodes = [(node if isinstance(node, Bound) else node.bound_node)
                for node in nodes]
    for node in nodes:
        self._modules[node.name] = node
        node.output_name = []
        if len(node.inputs) == 0:
            self.root_names.append(node.name)
    for node in nodes:
        for l_pre in node.inputs:
            l_pre.output_name.append(node.name)
        if (getattr(node, 'has_constraint', False) and
                node.name not in self.layers_with_constraint):
            self.layers_with_constraint.append(node.name)


def add_input_node(self: 'BoundedModule', node, index=None):
    self.add_nodes([node])
    self.input_name.append(node.name)
    # default value for input_index
    if index == 'auto':
        index = max([0] + [(i + 1)
                    for i in self.input_index if i is not None])
    self.input_index.append(index)


def delete_node(self: 'BoundedModule', node):
    for node_inp in node.inputs:
        node_inp.output_name.pop(node_inp.output_name.index(node.name))
    self._modules.pop(node.name)
    # TODO Create a list to contain all such lists such as
    # "relus" and "optimizable_activations"
    self.relus = [
        item for item in self.relus if item != node]
    self.optimizable_activations = [
        item for item in self.optimizable_activations if item != node]


def replace_node(self: 'BoundedModule', node_old, node_new):
    assert node_old != node_new
    for node in self.nodes():
        for i in range(len(node.inputs)):
            if node.inputs[i] == node_old:
                node.inputs[i] = node_new
    node_new.output_name += node_old.output_name
    if self.final_name == node_old.name:
        self.final_name = node_new.name
    for i in range(len(self.output_name)):
        if self.output_name[i] == node_old.name:
            self.output_name[i] = node_new.name
    self.delete_node(node_old)



"""Element multiplication with the A matrix based on its sign."""
import torch
from typing import Optional, Tuple
from torch import Tensor
from ..patches import Patches


torch._C._jit_set_profiling_executor(False)
torch._C._jit_set_profiling_mode(False)


class ClampedMultiplication(torch.autograd.Function):
    @staticmethod
    @torch.no_grad()
    # @torch.compile
    @torch.jit.script
    def clamp_mutiply_forward(A: Tensor, d_pos: Tensor, d_neg: Tensor,
            b_pos: Optional[Tensor], b_neg: Optional[Tensor], patches_mode: bool,
            reduce_bias: bool = False
        ) -> Tuple[Tensor, Tensor]:
        """Forward operations; actually the same as the reference implementation."""
        A_pos = A.clamp(min=0)
        A_neg = A.clamp(max=0)
        A_new = d_pos * A_pos + d_neg * A_neg
        bias_pos = bias_neg = torch.zeros(
                (), dtype=A_new.dtype, device=A_new.device)
        if b_pos is not None:
            if not reduce_bias:
                bias_pos = A_pos * b_pos
            else:
                if patches_mode:
                    bias_pos = torch.einsum('sb...chw,sb...chw->sb...', A_pos, b_pos)
                else:
                    bias_pos = torch.einsum('sb...,sb...->sb', A_pos, b_pos)
        if b_neg is not None:
            if not reduce_bias:
                bias_neg = A_neg * b_neg
            else:
                if patches_mode:
                    bias_neg = torch.einsum('sb...chw,sb...chw->sb...', A_neg, b_neg)
                else:
                    bias_neg = torch.einsum('sb...,sb...->sb', A_neg, b_neg)
        return A_new, bias_pos + bias_neg

    @staticmethod
    @torch.no_grad()
    # @torch.compile
    @torch.jit.script
    def clamp_mutiply_backward(A: Tensor, d_pos: Tensor, d_neg: Tensor,
            b_pos: Optional[Tensor], b_neg: Optional[Tensor],
            grad_output_A: Tensor, grad_output_bias: Optional[Tensor],
        ) -> Tuple[Tensor, Tensor, Tensor, Optional[Tensor], Optional[Tensor],
                   None, None]:
        """Improved backward operation. This should be better than the backward
        function generated by Pytorch."""
        if grad_output_bias is not None:
            extension_dim = len(A.shape) - len(grad_output_bias.shape)
            grad_output_bias = grad_output_bias.view(
                grad_output_bias.shape + (1, ) * extension_dim)
        A_pos_mask = (A >= 0).to(dtype=grad_output_A.dtype)
        A_neg_mask = 1. - A_pos_mask
        A_pos_grad_output_A = A_pos_mask * grad_output_A
        A_neg_grad_output_A = A_neg_mask * grad_output_A
        gd_pos = A * A_pos_grad_output_A
        gd_neg = A * A_neg_grad_output_A
        if b_pos is not None and b_neg is not None and grad_output_bias is not None:
            A_pos_grad_output_bias = A_pos_mask * grad_output_bias
            A_neg_grad_output_bias = A_neg_mask * grad_output_bias
            gb_neg = A * A_neg_grad_output_bias
            gb_pos = A * A_pos_grad_output_bias
            # gA has 4 terms.
            gA = (d_pos * A_pos_grad_output_A
                  + d_neg * A_neg_grad_output_A
                  + b_pos * A_pos_grad_output_bias
                  + b_neg * A_neg_grad_output_bias)
        elif b_neg is not None and grad_output_bias is not None:
            A_neg_grad_output_bias = A_neg_mask * grad_output_bias
            gb_neg = A * A_neg_grad_output_bias
            gb_pos = None
            # gA has 3 terms.
            gA = (d_pos * A_pos_grad_output_A
                  + d_neg * A_neg_grad_output_A
                  + b_neg * A_neg_grad_output_bias)
        elif b_pos is not None and grad_output_bias is not None:
            A_pos_grad_output_bias = A_pos_mask * grad_output_bias
            gb_pos = A * A_pos_grad_output_bias
            gb_neg = None
            # gA has 3 terms.
            gA = (d_pos * A_pos_grad_output_A + d_neg * A_neg_grad_output_A
                  + b_pos * A_pos_grad_output_bias)
        else:
            # gA has 2 terms.
            gA = d_pos * A_pos_grad_output_A + d_neg * A_neg_grad_output_A
            gb_pos = gb_neg = None
        return gA, gd_pos, gd_neg, gb_pos, gb_neg, None, None

    @staticmethod
    def forward(ctx, A, d_pos, d_neg, b_pos, b_neg, patches_mode, reduce_bias=True):
        # No need to save the intermediate A_pos, A_neg as they have been fused into the computation.
        ctx.save_for_backward(A, d_pos, d_neg, b_pos, b_neg)
        ctx.patches_mode = patches_mode
        ctx.reduce_bias = reduce_bias
        return ClampedMultiplication.clamp_mutiply_forward(
            A, d_pos, d_neg, b_pos, b_neg, patches_mode, reduce_bias)

    @staticmethod
    def backward(ctx, grad_output_A, grad_output_bias):
        A, d_pos, d_neg, b_pos, b_neg = ctx.saved_tensors
        assert ctx.reduce_bias
        return ClampedMultiplication.clamp_mutiply_backward(
            A, d_pos, d_neg, b_pos, b_neg,
            grad_output_A, grad_output_bias)


def multiply_by_A_signs(A, d_pos, d_neg, b_pos, b_neg, contiguous=True, reduce_bias=True):
    if isinstance(A, Tensor):
        if contiguous:
            # For dense mode, convert d_pos and d_neg to contiguous tensor by default.
            d_pos = d_pos.contiguous()
            d_neg = d_neg.contiguous()
        if d_pos.ndim == 1:
            # Special case for LSTM, the bias term is 1-dimension. (FIXME)
            assert d_neg.ndim == 1 and b_pos.ndim == 1 and b_neg.ndim == 1
            new_A = A.clamp(min=0) * d_pos + A.clamp(max=0) * d_neg
            new_bias = A.clamp(min=0) * b_pos + A.clamp(max=0) * b_neg
            return new_A, new_bias
        return ClampedMultiplication.apply(
            A, d_pos, d_neg, b_pos, b_neg, False, reduce_bias)
    elif isinstance(A, Patches):
        if contiguous:
            # For patches mode, do not convert d_pos and d_neg to contiguous tensor by default.
            d_pos = d_pos.contiguous()
            d_neg = d_neg.contiguous()
        assert A.identity == 0  # TODO: handle the A.identity = 1 case. Currently not used.
        patches = A.patches
        patches_shape = patches.shape
        # patches shape: [out_c, batch_size, out_h, out_w, in_c, H, W]. Here out_c is the spec dimension.
        # or (unstable_size, batch_size, in_c, H, W) when it is sparse.
        if len(patches_shape) == 6:
            patches = patches.view(*patches_shape[:2], -1, *patches_shape[-2:])
            d_pos = d_pos.view(*patches_shape[:2], -1, *patches_shape[-2:]) if d_pos is not None else None
            d_neg = d_neg.view(*patches_shape[:2], -1, *patches_shape[-2:]) if d_neg is not None else None
            b_pos = b_pos.view(*patches_shape[:2], -1, *patches_shape[-2:]) if b_pos is not None else None
            b_neg = b_neg.view(*patches_shape[:2], -1, *patches_shape[-2:]) if b_neg is not None else None
        # Apply the multiplication based on signs.
        A_prod, bias = ClampedMultiplication.apply(
            patches, d_pos, d_neg, b_pos, b_neg, True, reduce_bias)
        # prod has shape [out_c, batch_size, out_h, out_w, in_c, H, W] or (unstable_size, batch_size, in_c, H, W) when it is sparse.
        # For sparse patches the return bias size is (unstable_size, batch).
        # For regular patches the return bias size is (spec, batch, out_h, out_w).
        if len(patches_shape) == 6:
            A_prod = A_prod.view(*patches_shape)
        return A.create_similar(A_prod), bias




import torch
import torch.nn as nn

class CrossEntropyWrapper(nn.Module):
    def __init__(self, model):
        super(CrossEntropyWrapper, self).__init__()
        self.model = model

    def forward(self, x, labels):
        y = self.model(x)
        logits = y - torch.gather(y, dim=-1, index=labels.unsqueeze(-1))
        return torch.exp(logits).sum(dim=-1, keepdim=True)

class CrossEntropyWrapperMultiInput(nn.Module):
    def __init__(self, model):
        super(CrossEntropyWrapperMultiInput, self).__init__()
        self.model = model

    def forward(self, labels, *x):
        y = self.model(*x)
        logits = y - torch.gather(y, dim=-1, index=labels.unsqueeze(-1))
        return torch.exp(logits).sum(dim=-1, keepdim=True)



from beartype import beartype
import torch

class TensorStorage:

    """
    Fast managed dynamic sized tensor storage
    """

    # @beartype
    def __init__(self: 'TensorStorage', full_shape: torch.Tensor,
                 initial_size: int = 1024, switching_size: int = 65536,
                 device: str = 'cpu', concat_dim: int = 0) -> None:

        if isinstance(full_shape, torch.Tensor):
            data = full_shape
            full_shape = data.shape
        else:
            data = None

        self.shape = list(full_shape)
        self.dtype = torch.get_default_dtype()
        self.device = device
        self.concat_dim = concat_dim
        self.num_used = 0
        self.switching_size = switching_size

        self._storage = self._allocate(initial_size)

        if data is not None:
            self.append(data)


    # @beartype
    def _allocate(self: 'TensorStorage', new_size: int) -> torch.Tensor:
        allocate_shape = self.shape.copy()
        allocate_shape[self.concat_dim] = new_size
        if self.device == 'cpu' and torch.cuda.is_available():
            # pin CPU memory if cuda is available
            return torch.empty(allocate_shape, dtype=self.dtype, device=self.device, pin_memory=True)
        else:
            return torch.empty(allocate_shape, dtype=self.dtype, device=self.device)


    # @beartype
    def _get_new_size(self: 'TensorStorage', request_size: int) -> int:
        """Compute new size of storage given the current request."""
        if self._storage.size(self.concat_dim) < self.switching_size:
            # exponential growth with small tensor
            return max(self._storage.size(self.concat_dim) * 2, self.num_used + request_size)
        # linear growth with big tensor
        return self._storage.size(self.concat_dim) + request_size * 32


    # @beartype
    @torch.no_grad()
    def append(self: 'TensorStorage', appended_tensor: torch.Tensor) -> 'TensorStorage':
        """Append a new tensor to the storage object."""
        if self.num_used + appended_tensor.size(self.concat_dim) > self._storage.size(self.concat_dim):
            # Reallocate a new tensor, copying the existing contents over.
            new_size = self._get_new_size(appended_tensor.size(self.concat_dim))
            new_tensor = self._allocate(new_size)
            new_tensor.narrow(dim=self.concat_dim, start=0, length=self.num_used).copy_(
                self._storage.narrow(dim=self.concat_dim, start=0, length=self.num_used))
            # And then remove the old storage object.
            del self._storage
            self._storage = new_tensor
        self._storage.narrow(self.concat_dim, self.num_used, appended_tensor.size(self.concat_dim)).copy_(appended_tensor)
        self.num_used += appended_tensor.size(self.concat_dim)
        return self


    # @beartype
    @torch.no_grad()
    def pop(self: 'TensorStorage', size: int) -> torch.Tensor:
        """Remove tensors with 'size' at the end of the storage."""
        size = max(min(size, self.num_used), 0)
        ret = self._storage.narrow(self.concat_dim, self.num_used - size, size)
        self.num_used -= size
        return ret


    def __getattr__(self, attr):
        """Proxy all tensor attributes."""
        return getattr(self._storage.narrow(self.concat_dim, 0, self.num_used), attr)


    def __getitem__(self, idx):
        return self._storage.narrow(self.concat_dim, 0, self.num_used)[idx]


    def __len__(self):
        return self.num_used


    def __sub__(self, o):
        return self._storage.narrow(self.concat_dim, 0, self.num_used) - o._storage.narrow(o.concat_dim, 0, o.num_used)



import torch
import numpy as np

from ..utils import get_activation_value


def debug_model_conversion(onnx_model, inputs, pred_act, node, rtol=1e-3, atol=1e-4):
    """Compare if the activations of pytorch are the same as from onnxruntime."""
    if not isinstance(inputs, list):
        raise TypeError("inputs should be in a list.")

    if not all(isinstance(x, np.ndarray) for x in inputs):
        inputs = [x.detach().cpu().numpy() for x in inputs]

    exp_act = get_activation_value(onnx_model, inputs, list(node.output))
    if isinstance(pred_act, list):
        assert len(exp_act) == len(pred_act)
        for a, b in zip(exp_act, pred_act):
            exp = torch.from_numpy(a).cpu()
            pred = b.cpu()
            assert torch.equal(torch.tensor(exp.shape), torch.tensor(pred.shape))
            assert torch.allclose(exp, pred, rtol=rtol, atol=atol)
    else:
        exp = torch.from_numpy(exp_act[0]).cpu()
        pred = pred_act.cpu()
        assert torch.equal(torch.tensor(exp.shape), torch.tensor(pred.shape))
        assert torch.allclose(exp, pred, rtol=rtol, atol=atol)



import numpy as np
import torch

class SpecVNNLIB:

    def __init__(self, spec):

        self.bounds, self.mat = spec

        self.prop_mat = None
        self.prop_rhs = None
        self.true_labels = None
        self.target_labels = None


    def extract(self):
        # print('preprocess vnnlib spec')
        assert len(self.mat) == 1
        prop_mat, prop_rhs = self.mat[0]

        if self.prop_mat is None:
            self.prop_mat = torch.tensor(prop_mat, dtype=self.dtype, device=self.device).unsqueeze(0)

        if self.prop_rhs is None:
            self.prop_rhs = torch.tensor(prop_rhs, dtype=self.dtype, device=self.device).unsqueeze(0)


        if self.true_labels is None or self.target_labels is None:
            true_labels, target_labels = [], []
            for m in prop_mat:
                true_label = np.where(m == 1)[-1]
                if len(true_label) != 0:
                    assert len(true_label) == 1
                    true_labels.append(true_label[0])
                else:
                    true_labels.append(None)

                target_label = np.where(m == -1)[-1]
                if len(target_label) != 0:
                    assert len(target_label) == 1
                    target_labels.append(target_label[0])
                else:
                    target_labels.append(None)

            self.true_labels = np.array([true_labels])
            self.target_labels = np.array([target_labels])

        # print('c   ', self.prop_mat)
        # print('rhs ', self.prop_rhs)
        # print('y   ', self.true_labels)
        # print('pidx', self.target_labels)
        # exit()

        return self.prop_mat, self.prop_rhs, self.true_labels, self.target_labels


    def check_solution(self, output):
        for lhs, rhs in self.mat:
            lhs = torch.tensor(lhs, dtype=self.dtype, device=output.device)
            rhs = torch.tensor(rhs, dtype=self.dtype, device=output.device)
            vec = lhs @ output.squeeze(0)
            if torch.all(vec <= rhs):
                return True
        return False



'''
vnnlib simple utilities

Stanley Bak
June 2021
'''

from beartype import beartype
from copy import deepcopy
from pathlib import Path
import numpy as np
import tqdm
import re

from ..misc.logger import logger

# @beartype
def read_statements(vnnlib_filename: Path):
    '''process vnnlib and return a list of strings (statements)

    useful to get rid of comments and blank lines and combine multi-line statements
    '''

    with open(vnnlib_filename, 'r') as f:
        lines = f.readlines()

    lines = [line.strip() for line in lines]
    assert len(lines) > 0

    # combine lines if case a single command spans multiple lines
    open_parentheses = 0
    statements = []
    current_statement = ''

    for line in lines:
        comment_index = line.find(';')

        if comment_index != -1:
            line = line[:comment_index].rstrip()

        if not line:
            continue

        new_open = line.count('(')
        new_close = line.count(')')

        open_parentheses += new_open - new_close

        assert open_parentheses >= 0, "mismatched parenthesis in vnnlib file"

        # add space
        current_statement += ' ' if current_statement else ''
        current_statement += line

        if open_parentheses == 0:
            statements.append(current_statement)
            current_statement = ''

    if current_statement:
        statements.append(current_statement)

    # remove repeated whitespace characters
    statements = [" ".join(s.split()) for s in statements]

    # remove space after '('
    statements = [s.replace('( ', '(') for s in statements]

    # remove space after ')'
    statements = [s.replace(') ', ')') for s in statements]

    return statements


def update_rv_tuple(rv_tuple, op, first, second, num_inputs, num_outputs):
    'update tuple from rv in read_vnnlib_simple, with the passed in constraint "(op first second)"'

    if first.startswith("X_"):
        # Input constraints
        index = int(first[2:])

        assert not second.startswith("X") and not second.startswith("Y"), \
            f"input constraints must be box ({op} {first} {second})"
        assert 0 <= index < num_inputs, print(index, num_inputs)

        limits = rv_tuple[0][index]

        if op == "<=":
            limits[1] = min(float(second), limits[1])
        else:
            limits[0] = max(float(second), limits[0])

        assert limits[0] <= limits[1], f"{first} range is empty: {limits}"

    else:
        # output constraint
        if op == ">=":
            # swap order if op is >=
            first, second = second, first

        row = [0.0] * num_outputs
        rhs = 0.0

        # assume op is <=
        if first.startswith("Y_") and second.startswith("Y_"):
            index1 = int(first[2:])
            index2 = int(second[2:])

            row[index1] = 1
            row[index2] = -1
        elif first.startswith("Y_"):
            index1 = int(first[2:])
            row[index1] = 1
            rhs = float(second)
        else:
            assert second.startswith("Y_")
            index2 = int(second[2:])
            row[index2] = -1
            rhs = -1 * float(first)

        mat, rhs_list = rv_tuple[1], rv_tuple[2]
        mat.append(row)
        rhs_list.append(rhs)


def make_input_box_dict(num_inputs):
    'make a dict for the input box'

    rv = {i: [-np.inf, np.inf] for i in range(num_inputs)}

    return rv


# @beartype
def read_vnnlib(vnnlib_filename: str, regression: bool = False) -> list:
    return _read_vnnlib(vnnlib_filename=vnnlib_filename, regression=regression, mismatch_input_output=True)

# @beartype
def _read_vnnlib(vnnlib_filename: str, regression: bool = False, mismatch_input_output: bool = False) -> list:
    '''process in a vnnlib file

    this is not a general parser, and assumes files are provided in a 'nice' format. Only a single disjunction
    is allowed

    output a list containing 2-tuples:
        1. input ranges (box), list of pairs for each input variable
        2. specification, provided as a list of pairs (mat, rhs), as in: mat * y <= rhs, where y is the output.
                          Each element in the list is a term in a disjunction for the specification.

    If regression=True, the specification is a regression problem rather than classification.

    Currently we support vnnlib loader with cache:
        1. For the first time loading, it will parse the entire file and generate a cache file with md5 code of original file into *.compiled.
        2. For the later loading, it will check *.compiled and see if the stored md5 matches the original one. If not, regeneration is needed for vnnlib changing cases. Otherwise return the cache file.
    '''
    vnnlib_filename = Path(vnnlib_filename)
    assert vnnlib_filename.is_file() and vnnlib_filename.suffix == ".vnnlib", vnnlib_filename

    # example: "(declare-const X_0 Real)"
    regex_declare = re.compile(r"^\(declare-const (X|Y)_(\S+) Real\)$")

    # comparison sub-expression
    # example: "(<= Y_0 Y_1)" or "(<= Y_0 10.5)"
    comparison_str = r"\((<=|>=) (\S+) (\S+)\)"

    # example: "(and (<= Y_0 Y_2)(<= Y_1 Y_2))"
    dnf_clause_str = r"\(and\s*(" + comparison_str + r")+\)"

    # example: "(assert (<= Y_0 Y_1))"
    regex_simple_assert = re.compile(r"^\(assert " + comparison_str + r"\)$")

    # disjunctive-normal-form
    # (assert (or (and (<= Y_3 Y_0)(<= Y_3 Y_1)(<= Y_3 Y_2))(and (<= Y_4 Y_0)(<= Y_4 Y_1)(<= Y_4 Y_2))))
    regex_dnf = re.compile(r"^\(assert \(or (" + dnf_clause_str + r")+\)\)$")

    lines = read_statements(vnnlib_filename)

    # Read lines to determine number of inputs and outputs
    num_inputs = num_outputs = 0
    if mismatch_input_output:
        # logger.info(f'[!] Mismatch in VNNLIB')
        for line in lines:
            num = re.findall(r'X_(\d+)', line)
            if len(num):
                num_inputs = max(num_inputs, int(num[0]) + 1)
            num = re.findall(r'Y_(\d+)', line)
            if len(num):
                num_outputs = max(num_outputs, int(num[0]) + 1)
    else:
        for line in lines:
            declare = regex_declare.findall(line)
            if len(declare) == 0:
                continue
            elif len(declare) > 1:
                raise ValueError(f'There cannot be more than one declaration in one line: {line}')
            else:
                declare = declare[0]
                if declare[0] == 'X':
                    num_inputs = max(num_inputs, int(declare[1]) + 1)
                elif declare[0] == 'Y':
                    num_outputs = max(num_outputs, int(declare[1]) + 1)
                else:
                    raise ValueError(f'Unknown declaration: {line}')

    logger.info(f'[!] VNNLIB: {num_inputs} inputs, {num_outputs} outputs')

    rv = []  # list of 3-tuples, (box-dict, mat, rhs)
    rv.append((make_input_box_dict(num_inputs), [], []))

    if regression:
        # declare x0; declare y0; single assert
        assert len(lines) == 3

    for line in lines:
        if len(regex_declare.findall(line)) > 0:
            continue

        groups = regex_simple_assert.findall(line)

        if groups:
            assert len(groups[0]) == 3, f"groups was {groups}: {line}"
            op, first, second = groups[0]

            for rv_tuple in rv:
                update_rv_tuple(rv_tuple, op, first, second, num_inputs, num_outputs)

            continue

        ################
        groups = regex_dnf.findall(line)
        if not groups:
            logger.info(f"[VNNLIB] Skipped parsing line: {line}.")
            continue

        tokens = line.replace("(", " ").replace(")", " ").split()
        tokens = tokens[2:]  # skip 'assert' and 'or'

        conjuncts = " ".join(tokens).split("and")[1:]

        if regression:
            cases = []
            for c  in conjuncts:
                c_ = c.split()
                if c_[6] == '<=':
                    cases.append((float(c_[2]), float(c_[5]), float(c_[8]), 'lower'))
                elif c_[6] == '>=':
                    cases.append((float(c_[2]), float(c_[5]), float(c_[8]), 'upper'))
                else:
                    print(c_[6])
                    raise NotImplementedError
            return cases

        old_rv = rv
        rv = []

        for rv_tuple in old_rv:
            if len(conjuncts) > 10:
                pbar = tqdm.tqdm(conjuncts)
            else:
                pbar = conjuncts

            for c in pbar:
                rv_tuple_copy = deepcopy(rv_tuple)
                rv.append(rv_tuple_copy)

                c_tokens = [s for s in c.split(" ") if len(s) > 0]

                count = len(c_tokens) // 3

                for i in range(count):
                    op, first, second = c_tokens[3 * i:3 * (i + 1)]

                    update_rv_tuple(rv_tuple_copy, op, first, second, num_inputs, num_outputs)

    # merge elements of rv with the same input spec
    merged_rv = {}

    for rv_tuple in rv:
        boxdict = rv_tuple[0]
        matrhs = (rv_tuple[1], rv_tuple[2])

        key = str(boxdict)  # merge based on string representation of input box... accurate enough for now

        if key in merged_rv:
            merged_rv[key][1].append(matrhs)
        else:
            merged_rv[key] = (boxdict, [matrhs])

    # finalize objects (convert dicts to lists and lists to np.array)
    final_rv = []

    for rv_tuple in merged_rv.values():
        box_dict = rv_tuple[0]

        box = []

        for d in range(num_inputs):
            r = box_dict[d]

            assert r[0] != -np.inf and r[1] != np.inf, f"input X_{d} was unbounded: {r}"
            box.append(r)

        spec_list = []

        for matrhs in rv_tuple[1]:
            mat = np.array(matrhs[0], dtype=float)
            rhs = np.array(matrhs[1], dtype=float)
            spec_list.append((mat, rhs))

        final_rv.append((box, spec_list))

    return final_rv



import torch
from torch import nn


class Scatter(nn.Module):
    def __init__(self, dim=0):
        self.dim = dim
        super().__init__()

    def forward(self, data: torch.Tensor, indices: torch.Tensor, updates: torch.Tensor):
        return torch.scatter(data, self.dim, indices, updates)




from pathlib import Path
import torch.nn as nn
import random
import torch
import time
import os

from verifier.verifier import Verifier 
from util.spec.read_vnnlib import read_vnnlib
from util.network.read_onnx import parse_onnx
from verifier.objective import Objective, DnfObjectives
from util.misc.logger import logger
from setting import Settings


def extract_instance(net_path, vnnlib_path):
    vnnlibs = read_vnnlib(vnnlib_path)
    model, input_shape, output_shape, is_nhwc = parse_onnx(net_path)
    
    # objective
    objectives = []
    for spec in vnnlibs:
        bounds = spec[0]
        for prop_i in spec[1]:
            objectives.append(Objective((bounds, prop_i)))
    objectives = DnfObjectives(objectives, input_shape=input_shape, is_nhwc=is_nhwc)

    return model, input_shape, objectives


    
def test_1():
    net_path = 'example/backup/motivation_example_159.onnx'
    vnnlib_path = 'example/backup/motivation_example_159.vnnlib'
    
    # net_path = 'example/onnx/mnist-net_256x2.onnx'
    # vnnlib_path = 'example/vnnlib/prop_1_0.05.vnnlib'
    
    device = 'cpu'
    logger.setLevel(1)
    # Settings.setup(args=None)
    Settings.setup_test()
    
    print(Settings)
    
    print('Running test with', net_path, vnnlib_path)
    model, input_shape, objectives = extract_instance(net_path, vnnlib_path)
    model.to(device)
    print(model)

    verifier = Verifier(
        net=model, 
        input_shape=input_shape, 
        batch=1000,
        device=device,
    )
    
    # stable, unstable, lbs, ubs = verifier.compute_stability(objectives)
    # print('stable:', stable)
    # print('unstable:', unstable)
    
    preconditions = [
        # {'/input': (torch.tensor([0]), torch.tensor([-1.]), torch.tensor([0.])), '/input.3': ([], [], [])},
        {'/input': (torch.tensor([0, 1]), torch.tensor([1., 1.]), torch.tensor([0., 0.])), '/input.3': (torch.tensor([0]), torch.tensor([1.]), torch.tensor([0.]))},
        {'/input': (torch.tensor([0, 1]), torch.tensor([ 1., -1.]), torch.tensor([0., 0.])), '/input.3': (torch.tensor([0, 1]), torch.tensor([-1.,  1.]), torch.tensor([0., 0.]))},
    ]
    # preconditions = []
    
    print(preconditions)
    
    verifier.verify(objectives, preconditions=preconditions)
    print('status:', verifier.status)
    print('unsat core:', verifier.get_unsat_core())
    
    # for c in verifier._get_learned_conflict_clauses():
    #     print(c)
    # print('lbs:', lbs)
    # print('ubs:', ubs)
    
    print(verifier.get_stats())


def test_2():
    net_path = 'example/mnist-net_256x2.onnx'
    vnnlib_path = Path('example/prop_1_0.03.vnnlib')
    device = 'cuda'

    print('\n\nRunning test with', net_path, vnnlib_path)
    
    # preconditions = [eval(l.replace('tensor', 'torch.tensor')) for l in open('log.txt').read().strip().split('\n')][:400]
    # print(preconditions)
    preconditions = []
    
    model, input_shape, objectives = extract_instance(net_path, vnnlib_path)
    model.to(device)
    
    Settings.setup(args=None)
    print(Settings)
    logger.setLevel(1)
    
    verifier = Verifier(
        net=model, 
        input_shape=input_shape, 
        batch=200,
        device=device,
    )
    
    
    status = verifier.verify(objectives, preconditions=preconditions)
    print('status:', status, verifier.status)
    print('unsat core')
    from pprint import pprint
    print(verifier.get_unsat_core())
    
    # for c in verifier._get_learned_conflict_clauses():
    #     print(c)

if __name__ == "__main__":
    test_1()


from pathlib import Path
import torch.nn as nn
import random
import torch
import time
import os

from verifier.verifier import Verifier 
from util.spec.read_vnnlib import read_vnnlib
from util.network.read_onnx import parse_onnx
from verifier.objective import Objective, DnfObjectives
from util.misc.logger import logger
from setting import Settings


class PaperNet(nn.Module):
    def __init__(self):
        super().__init__()
        self.layers = nn.Sequential(*[
            nn.Linear(2, 2),
            nn.ReLU(),
            nn.Linear(2, 2),
            nn.ReLU(),
            nn.Linear(2, 2),
        ])
        self._init_weights()
        
    def random(self, size):
        while True:
            w = (torch.FloatTensor(*size).uniform_(-1, 1) * 10).int().float() / 10
            if (w != 0).all():
                return w
            
    def _init_weights(self):
        self.layers[0].weight.data = self.random(size=(2, 2))
        self.layers[0].bias.data = self.random(size=(2,))
        
        self.layers[2].weight.data = self.random(size=(2, 2))
        self.layers[2].bias.data = self.random(size=(2,))
        
        self.layers[4].weight.data = self.random(size=(2, 2))
        self.layers[4].bias.data = self.random(size=(2,))

    @torch.no_grad()
    def forward(self, x):
        return self.layers(x)

    def print_w_b(self):
        for layer in self.layers:
            if hasattr(layer, 'weight'):
                print(layer)
                print('\t[+] w:', layer.weight.data.flatten())
                print('\t[+] b:', layer.bias.data.flatten())
                print()


def generate_network(seed):
    # seed = 79616
    if seed is not None:
        torch.manual_seed(seed)
    net = PaperNet()
    x = torch.tensor([[1, 2]]).float()
    net.print_w_b()
    # print(net(x))
    
    net.eval()
    root_dir = '/home/droars/Desktop/neuralsat-priv/neuralsat/example'
    output_name = f'{root_dir}/motivation_example.onnx' #'example/cacmodel.onnx'
    os.system(f'rm -rf {output_name}')
    
    torch.onnx.export(
        net,
        x,
        output_name,
        verbose=False,
        opset_version=12,
    )
    
    assert os.path.exists(output_name)
    

def extract_instance(net_path, vnnlib_path):
    vnnlibs = read_vnnlib(Path(vnnlib_path))
    model, input_shape, output_shape, is_nhwc = parse_onnx(net_path)
    
    # objective
    objectives = []
    for spec in vnnlibs:
        bounds = spec[0]
        for prop_i in spec[1]:
            objectives.append(Objective((bounds, prop_i)))
    objectives = DnfObjectives(objectives, input_shape=input_shape, is_nhwc=is_nhwc)

    return model, input_shape, objectives


def verify():
    net_path = '/home/droars/Desktop/neuralsat-priv/neuralsat/example/motivation_example.onnx'
    vnnlib_path = '/home/droars/Desktop/neuralsat-priv/neuralsat/example/motivation_example.vnnlib'
    device = 'cpu'
    
    print('Running test with', net_path, vnnlib_path)
    model, input_shape, objectives = extract_instance(net_path, vnnlib_path)
    model.to(device)

    verifier = Verifier(
        net=model, 
        input_shape=input_shape, 
        batch=1000,
        device=device,
    )
    
    status = verifier.verify(objectives)
    print(status, verifier.iteration)
    return status, verifier.iteration

def save_network(seed):
    net_path = '/home/droars/Desktop/neuralsat-priv/neuralsat/example/motivation_example.onnx'
    new_net_path = f'/home/droars/Desktop/neuralsat-priv/neuralsat/example/motivation_example_{seed}.onnx'
    
    vnnlib_path = '/home/droars/Desktop/neuralsat-priv/neuralsat/example/motivation_example.vnnlib'
    new_vnnlib_path = f'/home/droars/Desktop/neuralsat-priv/neuralsat/example/motivation_example_{seed}.vnnlib'
    
    os.system(f'cp {net_path} {new_net_path}')
    os.system(f'cp {vnnlib_path} {new_vnnlib_path}')
    

if __name__ == "__main__":
    seed = None
    Settings.setup_test()
    logger.setLevel(2)
    
    for trial in range(300):
        print('\n\n[+] Trail:', trial)
        status, iteration = None, 0
            
        # seed = random.randint(0, 100000)
        # seed = 98556
        # print('seed:', seed)
        try:
            generate_network(seed)
            status, iteration = verify()
        except:
            # raise
            print('[!] Error occurred')
            pass
        
        if status == 'unsat' and iteration > 2:
            save_network(trial)
            # break
        
        time.sleep(1)


import torch
from .bound_ops import *
from .utils import logger

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .bound_general import BoundedModule


def IBP_general(self: 'BoundedModule', node=None, C=None,
                delete_bounds_after_use=False):

    logger.debug('IBP for %s', node)

    def _delete_unused_bounds(node_list):
        """Delete bounds from input layers after use to save memory. Used when
        sparse_intermediate_bounds_with_ibp is true."""
        if delete_bounds_after_use:
            for n in node_list:
                del n.interval
                del n.lower
                del n.upper

    if self.bound_opts.get('loss_fusion', False):
        res = self._IBP_loss_fusion(node, C)
        if res is not None:
            return res

    if not node.perturbed:
        fv = self.get_forward_value(node)
        node.lower, node.upper = node.interval = (fv, fv)

    to_be_deleted_bounds = []
    if not hasattr(node, 'interval'):
        for n in node.inputs:
            if not hasattr(n, 'interval'):
                # Node n does not have interval bounds; we must compute it.
                self.IBP_general(
                    n, delete_bounds_after_use=delete_bounds_after_use)
                to_be_deleted_bounds.append(n)
        inp = [n_pre.interval for n_pre in node.inputs]
        if (C is not None and isinstance(node, BoundLinear)
                and not node.is_input_perturbed(1)):
            # merge the last BoundLinear node with the specification, available
            # when weights of this layer are not perturbed
            ret = node.interval_propagate(*inp, C=C)
            _delete_unused_bounds(to_be_deleted_bounds)
            return ret
        else:
            node.interval = node.interval_propagate(*inp)

        node.lower, node.upper = node.interval
        if isinstance(node.lower, torch.Size):
            node.lower = torch.tensor(node.lower)
            node.interval = (node.lower, node.upper)
        if isinstance(node.upper, torch.Size):
            node.upper = torch.tensor(node.upper)
            node.interval = (node.lower, node.upper)

    if C is not None:
        _delete_unused_bounds(to_be_deleted_bounds)
        return BoundLinear.interval_propagate(None, node.interval, C=C)
    else:
        _delete_unused_bounds(to_be_deleted_bounds)
        return node.interval

def _IBP_loss_fusion(self: 'BoundedModule', node, C):
    """Merge BoundLinear, BoundGatherElements and BoundSub.

    Improvement when loss fusion is used in training.
    """
    # not using loss fusion
    if not self.bound_opts.get('loss_fusion', False):
        return None

    # Currently this function has issues in more complicated networks.
    if self.bound_opts.get('no_ibp_loss_fusion', False):
        return None

    if (C is None and isinstance(node, BoundSub)
            and isinstance(node.inputs[1], BoundGatherElements)
            and isinstance(node.inputs[0], BoundLinear)):
        node_gather = node.inputs[1]
        node_linear = node.inputs[0]
        node_start = node_linear.inputs[0]
        w = node_linear.inputs[1].param
        b = node_linear.inputs[2].param
        labels = node_gather.inputs[1]
        if not hasattr(node_start, 'interval'):
            self.IBP_general(node_start)
        for n in node_gather.inputs:
            if not hasattr(n, 'interval'):
                self.IBP_general(n)
        if torch.isclose(labels.lower, labels.upper, 1e-8).all():
            labels = labels.lower
            batch_size = labels.shape[0]
            w = w.expand(batch_size, *w.shape)
            w = w - torch.gather(
                w, dim=1,
                index=labels.unsqueeze(-1).repeat(1, w.shape[1], w.shape[2]))
            b = b.expand(batch_size, *b.shape)
            b = b - torch.gather(b, dim=1,
                                    index=labels.repeat(1, b.shape[1]))
            lower, upper = node_start.interval
            lower, upper = lower.unsqueeze(1), upper.unsqueeze(1)
            node.lower, node.upper = node_linear.interval_propagate(
                (lower, upper), (w, w), (b.unsqueeze(1), b.unsqueeze(1)))
            node.interval = node.lower, node.upper = (
                node.lower.squeeze(1), node.upper.squeeze(1))
            return node.interval

    return None


def check_IBP_intermediate(self: 'BoundedModule', node):
    """ Check if we use IBP bounds to compute intermediate bounds on this node.

    Currently, assume all eligible operators have exactly one input.
    """
    if (isinstance(node, BoundReshape)
            and hasattr(node.inputs[0], 'lower')
            and hasattr(node.inputs[1], 'value')):
        # Node for input value.
        val_input = node.inputs[0]
        # Node for input parameter (e.g., shape, permute)
        arg_input = node.inputs[1]
        node.lower = node.forward(val_input.lower, arg_input.value)
        node.upper = node.forward(val_input.upper, arg_input.value)
        node.interval = (node.lower, node.upper)
        return True

    nodes = []
    while (getattr(node, 'lower', None) is None
            or getattr(node, 'upper', None) is None):
        if not node.ibp_intermediate:
            return False
        assert len(node.inputs) == 1, (
            'Nodes with ibp_intermediate=True cannot have more than one input')
        nodes.append(node)
        node = node.inputs[0]  # FIXME: this cannot handle multiple inputs.
    nodes.reverse()
    for n in nodes:
        n.interval = self.IBP_general(n)

    return True


def check_IBP_first_linear(self: 'BoundedModule', node):
    """Here we avoid creating a big C matrix in the first linear layer.
    Disable this optimization when we have beta for intermediate layer bounds.
    Disable this optimization when we need the A matrix of the first nonlinear
    layer, forcibly use CROWN to record A matrix.
    """
    # This is the list of all intermediate layers where we need to refine.
    if self.intermediate_constr is not None:
        intermediate_beta_enabled_layers = [
            k for v in self.intermediate_constr.values() for k in v]
    else:
        intermediate_beta_enabled_layers = []

    if (node.name not in self.needed_A_dict.keys()
            and (type(node) == BoundLinear
                or type(node) == BoundConv
                and node.name not in intermediate_beta_enabled_layers)):
        if type(node.inputs[0]) == BoundInput:
            node.lower, node.upper = self.IBP_general(node)
            return True

    return False



import time

import numpy as np
from tqdm import tqdm


class tqdmm(tqdm):

    def __init__(self, *args, **kwargs):
        if not len(args) and not len(kwargs):
            kwargs.update(total=100)
        super().__init__(*args, **kwargs)

    def set_n(self, n):
        self.update(n - self.n)

    def set_percentage(self, p):
        if p > 1:
            raise ValueError(f'p must be in range [0, 1]. Got {p}.')
        self.total = 100
        self.set_n(round(p * 100))


def process(items):
    new_items = [None] * len(items) * 2
    base_p = np.clip(np.random.normal(0.45, 0.2), 0., 1.)
    valid_mask = np.random.binomial(1, p=np.full(100, base_p))
    valid_items = [new_items[i] for i in range(len(new_items)) if valid_mask[i]]
    return valid_items


def main(init_size=200, batch_size=2):
    items = [None] * init_size

    pbar = tqdmm()
    visited = 0
    while True:
        items, to_be_processed_items = items[:-batch_size], items[-batch_size:]
        new_items = process(to_be_processed_items)
        items.extend(new_items)
        visited += 2 * len(to_be_processed_items)

        time.sleep(0.1)
        pbar.set_percentage(1 - min(1, (len(items) / visited)))
        pbar.set_description(f'n_items={len(items)}')
        if len(items) == 0:
            break
    pbar.close()


if __name__ == '__main__':
    main()



""" Resize operator """
import itertools

import torch

from .base import *
import numpy as np
from .solver_utils import grb
from ..patches import unify_shape, create_valid_mask, is_shape_used
from .gradient_modules import Conv2dGrad


class BoundResize(Bound):
    def __init__(self, attr=None, inputs=None, output_index=0, options=None):
        super().__init__(attr, inputs, output_index, options)
        # only support nearest mode for now
        assert attr["mode"] == "nearest"
        self.mode = attr["mode"]
        self.scale_factor = None

    def forward(self, x, size=None, scale_factor=None):
        # currently, forwarding size is not supported.
        assert isinstance(size, torch.Tensor) and len(size.tolist()) == 0
        # currently, only support enlarge tensor size by an integer factor.
        assert len(scale_factor.tolist()) == 4 and np.array([tmp.is_integer() and tmp > 0 for tmp in scale_factor.tolist()]).all()
        assert (scale_factor[0:2].to(torch.long) == 1).all(), 'only support resize on the H and W dim'
        self.scale_factor = tuple([int(tmp) for tmp in scale_factor][2:])
        if x.ndim == 4:
            final = F.interpolate(
                x, None, self.scale_factor, mode=self.mode)
        else:
            raise NotImplementedError(
                "Interpolation in 3D or interpolation with parameter size has not been implmented.")
        return final

    def interval_propagate(self, *v):
        l, u = zip(*v)
        return Interval.make_interval(self.forward(*l), self.forward(*u), v[0])

    def bound_forward(self, dim_in, *inp):
        x = inp[0]
        lw, lb, uw, ub = x.lw, x.lb, x.uw, x.ub
        new_lw, new_lb, new_uw, new_ub = \
            torch.nn.functional.upsample(lw, scale_factor=([1] * (lw.ndim - 4)) + list(self.scale_factor), mode=self.mode), \
            torch.nn.functional.upsample(lb, scale_factor=([1] * (lb.ndim - 4)) + list(self.scale_factor), mode=self.mode), \
            torch.nn.functional.upsample(uw, scale_factor=([1] * (uw.ndim - 4)) + list(self.scale_factor), mode=self.mode), \
            torch.nn.functional.upsample(ub, scale_factor=([1] * (ub.ndim - 4)) + list(self.scale_factor), mode=self.mode)
        return LinearBound(
            lw = new_lw,
            lb = new_lb,
            uw = new_uw,
            ub = new_ub)

    def bound_backward(self, last_lA, last_uA, *x, **kwargs):

        def _bound_oneside(last_A):
            if last_A is None:
                return None
            assert type(last_A) is Patches or last_A.ndim == 5
            # in case the kernel size cannot be divided by scale_factor, we round up the shape
            split_shape = tuple((torch.tensor(
                last_A.shape)[-2:] / torch.tensor(self.scale_factor)).ceil().to(torch.long).tolist())
            new_shape = last_A.shape[:-2] + split_shape
            if not type(last_A) is Patches:
                # classical mode is simple to handle by
                # sum the grid elements by using avg_pool2d with divisor_override=1
                return torch.nn.functional.avg_pool2d(
                    last_A.reshape(-1, *last_A.shape[-2:]), kernel_size=self.scale_factor, stride=self.scale_factor,
                    divisor_override=1).reshape(new_shape)
            else:
                # for patches mode
                assert type(last_A) is Patches
                assert self.scale_factor[0] == self.scale_factor[1]
                if self.scale_factor[0] == 1:
                    # identity upsampling
                    return last_A
                if isinstance(last_A.padding, int) and last_A.padding % self.scale_factor[0] == 0 and last_A.stride % self.scale_factor[0] == 0 and last_A.inserted_zeros == 0:
                    # an easy case where patch sliding windows coincides with the nearest sampling scaling windows
                    # in this case, we divide each patch to size of scale_factor sub-matrices,
                    # and sum up each sub-matrices respectively
                    # print(last_A.shape)
                    padding = last_A.shape[-1] % self.scale_factor[-1]
                    new_patches = torch.nn.functional.pad(last_A.patches, (0, padding, 0, padding))
                    new_patches = torch.nn.functional.avg_pool2d(
                        new_patches.reshape(-1, *new_patches.shape[-2:]), kernel_size=self.scale_factor,
                        stride=self.scale_factor, divisor_override=1).reshape(new_shape)
                    return last_A.create_similar(patches=new_patches,
                                                 stride=last_A.stride//self.scale_factor[0],
                                                 padding=last_A.padding//self.scale_factor[0],
                                                 )
                else:
                    """
                        The following part is created and mainly maintained by Linyi
                        Time complexity = O(A.numel * scale_factor + outH * kerH + outW * kerW + A.numel * kerH * kerW)
                        With Python loop complexity = O(outH + outW + kerH * kerW * scale_factor^2)
                    """
                    # preparation: unify shape
                    if last_A.padding:
                        padding = unify_shape(last_A.padding)
                    else:
                        padding = (0,0,0,0)
                    # padding = (left, right, top, bottom)
                    if last_A.output_padding:
                        output_padding = unify_shape(last_A.output_padding)
                    else:
                        output_padding = (0,0,0,0)
                    # output_padding = (left, right, top, bottom)

                    """
                        Step 0: filter out valid entries that maps to real cells of input
                        Like with inserted zeros = 2, [x 0 0 x 0 0 x]. Only "x" cells are kept
                        Borrowed from one_d generation from Conv patches
                    """
                    one_d_unfolded_r = create_valid_mask(self.output_shape,
                                                         last_A.patches.device,
                                                         last_A.patches.dtype,
                                                         last_A.patches.shape[-2:],
                                                         last_A.stride,
                                                         last_A.inserted_zeros,
                                                         last_A.padding,
                                                         last_A.output_padding,
                                                         last_A.unstable_idx)
                    patches = last_A.patches * one_d_unfolded_r

                    """
                        Step 1: compute the coordinate mapping from patch coordinates to input coordinates
                        Time complexity: O(outH + outW)
                        note: last_A shape is [outC, batch, outH, outW, inC, kerH, kerW]
                        We create H_idx_map and W_idx_map of shape [outH] and [outW] respectively,
                        recording the start idx of row/column for patches at position [.,.,.,.,.,i,j]
                        in H_idx_map[i] and W_idx_map[j]
                    """
                    ker_size_h, ker_size_w = last_A.shape[-2], last_A.shape[-1]
                    if last_A.unstable_idx is None:
                        # we can get the real output H and W from shape[2] and shape [3]
                        out_h, out_w = last_A.shape[2], last_A.shape[3]
                    else:
                        # it seems to be stored in output_shape
                        out_h, out_w = last_A.output_shape[-2], last_A.output_shape[-1]
                    h_idx_map = torch.arange(0, out_h) * last_A.stride - padding[-2] + output_padding[-2] * last_A.stride
                    h_idx_map = h_idx_map.to(last_A.device)
                    w_idx_map = torch.arange(0, out_w) * last_A.stride - padding[-4] + output_padding[-4] * last_A.stride
                    w_idx_map = w_idx_map.to(last_A.device)

                    r"""
                        Step 2: compute the compressed patches
                        Time complexity: O(outH * kerH + outW * kerW + A.numel * kerH * kerW)
                        Upsampling needs to sum up A cells in scale_factor * scale_factor sub-blocks
                        Example: when scale factor is 2
                        [ a b c d
                          e f g h    ---\    [ a+b+e+f c+d+g+h
                          i j k l    ---/      i+j+m+n k+l+o+p]
                          m n o p]
                        In patches mode, we need to sum up cells in each patch accordingly.
                        The summing mechanism could change at different locations.
                        For each spatial dimension, we create a binary sum_mask tensor [outH, ker_size_h, new_ker_size_h]
                            to select the cells to sum up
                        Example:
                        For [a b c d] -> [a+b c+d], with 3x3 patch covering [0..2] and [2..4].
                        The first patch needs to sum to [a+b c]; the second patch needs to sum to [b c+d]
                        So we have sum_mask
                        [ for patch 1: [[1, 1, 0],    (first entry sums up index 0 and 1)
                                        [0, 0, 1]]^T, (second entry sums up index 2)
                          for patch 2: [[1, 0, 0],    (first entry sums up index 0)
                                        [0, 1, 1]]^T  (second entry sums up index 1 and 2)
                        ]
                        With the mask, we can now compute the new patches with einsum:
                            [outC, batch, outH, outW, inC, kerH, kerW] * [outH, kerH, new_kerH] -> [outC, batch, outH, outW, inC, new_kerH, kerW]
                    """
                    tot_scale_fac = ((last_A.inserted_zeros + 1) * self.scale_factor[0], (last_A.inserted_zeros + 1) * self.scale_factor[1])
                    new_ker_size_h, new_ker_size_w = \
                        (tot_scale_fac[0] + ker_size_h - 2) // tot_scale_fac[0] + 1, \
                        (tot_scale_fac[1] + ker_size_w - 2) // tot_scale_fac[1] + 1

                    min_h_idx, max_h_idx = h_idx_map[0], h_idx_map[-1] + ker_size_h
                    shrank_h_idx = (torch.arange(min_h_idx, max_h_idx) + last_A.inserted_zeros).div(tot_scale_fac[0], rounding_mode='floor')
                    if last_A.unstable_idx is None:
                        # with nonsparse index, create full-sized sum musk for rows
                        ker_h_indexer = torch.arange(0, ker_size_h).to(last_A.device)
                        sum_mask_h = torch.zeros(last_A.shape[2], ker_size_h, new_ker_size_h).to(last_A.device)
                        for i in range(last_A.shape[2]):
                            sum_mask_h[i, ker_h_indexer, \
                                shrank_h_idx[h_idx_map[i] - min_h_idx: h_idx_map[i] - min_h_idx + ker_size_h] - shrank_h_idx[h_idx_map[i] - min_h_idx]] = 1
                            # set zero to those in padding area
                            padding_place_mask = (ker_h_indexer + h_idx_map[i] < 0)
                            sum_mask_h[i, padding_place_mask] = 0
                    else:
                        # with sparse index, create sparse sum musk
                        sum_mask_h = torch.zeros(last_A.shape[0], ker_size_h, new_ker_size_h).to(last_A.device)

                        row_nos = last_A.unstable_idx[1]
                        unstable_loc_indexer = torch.arange(0, row_nos.shape[0]).to(last_A.device)

                        for k in range(ker_size_h):
                            place_in_new_ker = shrank_h_idx[h_idx_map[row_nos] - min_h_idx + k] - shrank_h_idx[h_idx_map[row_nos] - min_h_idx]
                            sum_mask_h[unstable_loc_indexer, k, place_in_new_ker] = 1
                            # set zero to those in padding area
                            padding_place_mask = (h_idx_map[row_nos] + k < 0)
                            sum_mask_h[padding_place_mask, k] = 0

                    min_w_idx, max_w_idx = w_idx_map[0], w_idx_map[-1] + ker_size_w
                    shrank_w_idx = (torch.arange(min_w_idx, max_w_idx) + last_A.inserted_zeros).div(tot_scale_fac[1], rounding_mode='floor')
                    if last_A.unstable_idx is None:
                        # with nonsparse index, create full-sized sum musk for columns
                        ker_w_indexer = torch.arange(0, ker_size_w).to(last_A.device)
                        sum_mask_w = torch.zeros(last_A.shape[3], ker_size_w, new_ker_size_w).to(last_A.device)
                        for i in range(last_A.shape[3]):
                            sum_mask_w[i, ker_w_indexer, \
                                shrank_w_idx[w_idx_map[i] - min_w_idx: w_idx_map[i] - min_w_idx + ker_size_w] - shrank_w_idx[w_idx_map[i] - min_w_idx]] = 1
                            # set zero to those in padding area
                            padding_place_mask = (ker_w_indexer + w_idx_map[i] < 0)
                            sum_mask_w[i, padding_place_mask] = 0
                    else:
                        # with sparse index, create sparse sum musk
                        sum_mask_w = torch.zeros(last_A.shape[0], ker_size_w, new_ker_size_w).to(last_A.device)

                        col_nos = last_A.unstable_idx[2]
                        unstable_loc_indexer = torch.arange(0, col_nos.shape[0]).to(last_A.device)

                        for k in range(ker_size_w):
                            place_in_new_ker = shrank_w_idx[w_idx_map[col_nos] - min_w_idx + k] - shrank_w_idx[w_idx_map[col_nos] - min_w_idx]
                            sum_mask_w[unstable_loc_indexer, k, place_in_new_ker] = 1
                            # set zero to those in padding area
                            padding_place_mask = (w_idx_map[col_nos] + k < 0)
                            sum_mask_w[padding_place_mask, k] = 0

                    if last_A.unstable_idx is None:
                        # nonsparse aggregation
                        new_patches = torch.einsum("ObhwIij,hix,wjy->ObhwIxy", patches, sum_mask_h, sum_mask_w)
                    else:
                        # sparse aggregation
                        new_patches = torch.einsum("NbIij,Nix,Njy->NbIxy", patches, sum_mask_h, sum_mask_w)

                    """
                        Step 3: broadcasting the new_patches by repeating elements,
                            since later we would need to apply insert_zeros
                        For example, scale_factor = 3, repeat patch [a,b] to [a,a,a,b,b,b]
                        Time complexity: O(A.numel * scale_factor)
                    """
                    ext_new_ker_size_h, ext_new_ker_size_w = \
                        new_ker_size_h * tot_scale_fac[0], new_ker_size_w * tot_scale_fac[1]
                    ext_new_patches = torch.zeros(list(new_patches.shape[:-2]) +
                                                  [ext_new_ker_size_h, ext_new_ker_size_w], device=new_patches.device)
                    for i in range(ext_new_ker_size_h):
                        for j in range(ext_new_ker_size_w):
                            ext_new_patches[..., i, j] = new_patches[..., i // tot_scale_fac[0], j // tot_scale_fac[1]]

                    """
                        Step 4: compute new padding, stride, shape, insert_zeros, and output_padding
                    """
                    # stride should be the same after upsampling, stride is an integer
                    # new_stride = last_A.stride
                    # padding can change much, the beginning should extend by (scale - 1) entries,
                    # the ending should extend by (ext_new_ker_size - ker_size) entries
                    # padding = (left, right, top, bottom)
                    new_padding = (padding[0] + (self.scale_factor[1] - 1) * (last_A.inserted_zeros + 1),
                                   padding[1] + ext_new_ker_size_w - ker_size_w,
                                   padding[2] + (self.scale_factor[0] - 1) * (last_A.inserted_zeros + 1),
                                   padding[3] + ext_new_ker_size_h - ker_size_h)
                    if new_padding[0] == new_padding[1] and new_padding[1] == new_padding[2] and new_padding[2] == new_padding[3]:
                        # simplify to an int
                        new_padding = new_padding[0]
                    # only support uniform scaling on H and W now, i.e., self.scale_factor[0] == self.scale_factor[1]
                    inserted_zeros = tot_scale_fac[0] - 1
                    # output padding seems not to change
                    # new_output_padding = last_A.output_padding

                    """
                        Package and create
                    """
                    # sparse tensor doesn't support einsum which is necessary for subsequent computes, so deprecated
                    # if inserted_zeros >= 3:
                    #     # mask unused cells
                    #     input_shape = list(self.output_shape)
                    #     input_shape[-2], input_shape[-1] = input_shape[-2] // self.scale_factor[-2], \
                    #         input_shape[-1] // self.scale_factor[-1]
                    #     one_unfolded = create_valid_mask(input_shape, ext_new_patches.device,
                    #                                       ext_new_patches.dtype, ext_new_patches.shape[-2:],
                    #                                       last_A.stride, inserted_zeros, new_padding,
                    #                                       last_A.output_padding,
                    #                                       last_A.unstable_idx if last_A.unstable_idx else None)
                    #     ext_new_patches = (ext_new_patches * one_unfolded).to_sparse()

                    # print the shape change after upsampling, if needed
                    # print(f'After upsampling, '
                    #       f'{last_A.patches.shape} (pad={padding}, iz={last_A.inserted_zeros}, s={last_A.stride}) -> '
                    #       f'{ext_new_patches.shape} (pad={new_padding}, iz={inserted_zeros}, s={last_A.stride})')
                    ret_patches_A = last_A.create_similar(patches=ext_new_patches,
                                                          padding=new_padding,
                                                          inserted_zeros=inserted_zeros)
                    if self.input_shape[-2] < ret_patches_A.shape[-2] and self.input_shape[-1] < ret_patches_A.shape[-2] \
                            and not is_shape_used(ret_patches_A.output_padding):
                        # using matrix mode could be more memory efficient
                        ret_matrix_A = ret_patches_A.to_matrix(self.input_shape)
                        # print(f'After upsampling, to_matrix: {ret_matrix_A.shape}')
                        ret_matrix_A = ret_matrix_A.transpose(0, 1)
                        return ret_matrix_A
                    else:
                        return ret_patches_A

        last_lA = _bound_oneside(last_lA)
        last_uA = _bound_oneside(last_uA)
        return [(last_lA, last_uA), (None, None), (None, None)], 0, 0



from beartype import beartype
import typing
import torch

from ..setting import Settings

# @beartype
def get_check_abstractor_params() -> dict:
    return {
        'crown_batch_size': Settings.backward_batch_size,
        'forward_max_dim': Settings.forward_max_dim,
        'dynamic_forward': Settings.forward_dynamic,
        'optimize_bound_args': {
            'iteration': 1,
            'stop_criterion_func': lambda x: False,
            'enable_beta_crown': False
        }
    }

# @beartype
def get_branching_opt_params() -> dict:
    return {
        'crown_batch_size': Settings.backward_batch_size,
        'optimize_bound_args': {
            'enable_beta_crown': False,
            'fix_interm_bounds': True,
        }
    }


# @beartype
def get_initialize_opt_params(stop_criterion_func: typing.Callable) -> dict:
    return {
        'crown_batch_size': Settings.backward_batch_size,
        'optimize_bound_args': {
            'enable_alpha_crown': True,
            'enable_beta_crown': False,
            'use_shared_alpha': Settings.share_alphas,
            'init_alpha': False,
            'fix_interm_bounds': True,
            'stop_criterion_func': stop_criterion_func,
            'iteration': 50,
            'lr_alpha': 0.1,
            'lr_beta': 0.1,
            'lr_decay': 0.98,
        }
    }


# @beartype
def get_beta_opt_params(stop_criterion_func: typing.Callable) -> dict:
    return {
        'crown_batch_size': Settings.backward_batch_size,
        'optimize_bound_args': {
            'enable_alpha_crown': True,
            'enable_beta_crown': True,
            'use_shared_alpha': Settings.share_alphas,
            'fix_interm_bounds': True,
            'iteration': 20,
            'lr_alpha': 0.1,
            'lr_beta': 0.1,
            'lr_decay': 0.98,
            'stop_criterion_func': stop_criterion_func,
        }
    }


# @beartype
def get_input_opt_params(stop_criterion_func: typing.Callable) -> dict:
    return {
        'crown_batch_size': Settings.backward_batch_size,
        'forward_max_dim': Settings.forward_max_dim,
        'dynamic_forward': Settings.forward_dynamic,
        'optimize_bound_args': {
            'enable_beta_crown': False,
            'fix_interm_bounds': True,
            'iteration': 20,
            'lr_alpha': 0.1,
            'lr_decay': 0.98,
            'stop_criterion_func': stop_criterion_func,
        }
    }



class DummyGurobipyClass:
    """A dummy class with error message when gurobi is not installed."""
    def __getattr__(self, attr):
        def _f(*args, **kwargs):
            raise RuntimeError(f"method {attr} not available because gurobipy module was not built.")
        return _f

try:
    import gurobipy as grb
except ModuleNotFoundError:
    grb = DummyGurobipyClass()


import torch
from torch import nn

from ..utils import PRINT_DEBUG

class Range(nn.Module):
    def forward(self, start: torch.Tensor, limit: torch.Tensor, delta: torch.Tensor):

        if PRINT_DEBUG:
            print('RANGE:', start, limit, delta)

        if start.numel() == 1:
            start = start.item()

        if limit.numel() == 1:
            limit = limit.item()

        if delta.numel() == 1:
            delta = delta.item()

        if PRINT_DEBUG:
            print('\t- start:', start)
            print('\t- limit:', limit)
            print('\t- delta:', delta)

        final = torch.arange(start=start, end=limit, step=delta)

        if PRINT_DEBUG:
            print('\t- final:', final)
            print('\t- final:', final.shape)

        return final



from abc import ABC

from torch import nn


class Operator(nn.Module, ABC):
    @staticmethod
    def get_axis(input_shape, input_feature_axis):
        """
        Parameters
        ----------
        input_shape: torch.Size
        input_feature_axis: int

        Returns
        -------
        axis: tuple
            Axis to aggregate over.
        """
        if input_feature_axis < 0:
            input_feature_axis += len(input_shape)
            # select and sum all axes except the feature one
        axis = set(range(len(input_shape))) - {input_feature_axis}
        return tuple(axis)


class OperatorWrapper(Operator, ABC):
    def __init__(self, op):
        """
        This class enables any function to become a subclass of nn.Module
        The class name is equal to the op.__name__

        Parameters
        ----------
        op: function or builtin_function_or_method
            Any torch function. It is used in-place of forward method.
        """
        self.forward = op
        self.__class__.__name__ = op.__name__
        super().__init__()



from .bound_general import BoundedModule
from .bound_multi_gpu import BoundDataParallel
from .bounded_tensor import BoundedTensor, BoundedParameter
from .perturbations import PerturbationLpNorm, PerturbationSynonym
from .wrapper import CrossEntropyWrapper, CrossEntropyWrapperMultiInput
from .bound_op_map import register_custom_op, unregister_custom_op

__version__ = '0.4.0'



import warnings
warnings.filterwarnings(action='ignore')

from beartype import beartype
import traceback
import logging
import typing
import torch
import copy
import math
import os

from ..auto_LiRPA.utils import stop_criterion_batch_any
from ..auto_LiRPA import BoundedModule
from ..onnx2pytorch.convert.model import ConvertModel
from ..util.misc.result import AbstractResults
from ..util.misc.logger import logger
from .params import *


class NetworkAbstractor:

    "Over-approximation method alpha-beta-CROWN"

    # @beartype
    def __init__(self: 'NetworkAbstractor', pytorch_model: ConvertModel, input_shape: tuple, method: str, input_split: bool = False, device: str = 'cpu') -> None:

        self.pytorch_model = copy.deepcopy(pytorch_model)
        self.device = device
        self.input_shape = input_shape

        # search domain
        self.input_split = input_split

        # computation algorithm
        self.method = method

        # debug
        self.iteration = 0

    # @beartype
    @property
    def split_points(self):
        if not hasattr(self, '_split_points'):
            self._split_points = [self.net.split_activations[k.name][0][0].get_split_point() for k in self.net.split_nodes]
        return self._split_points

    # @beartype
    def setup(self: 'NetworkAbstractor', objective: typing.Any) -> None:
        if self.select_params(objective):
            logger.info(f'Initialized abstractor: mode="{self.mode}", method="{self.method}", input_split={self.input_split}, backward_batch_size={Settings.backward_batch_size}')
            return None

        # FIXME: try special settings for ViT
        extra_opts = {'sparse_intermediate_bounds': False, 'buffers': {'no_batchdim': True},}
        if self.select_params(objective, extra_opts=extra_opts):
            logger.info(f'Initialized abstractor: mode="{self.mode}", method="{self.method}", input_split={self.input_split}, extra_opts={extra_opts}')
            return None

        # FIXME: try smaller backward batch size
        Settings.backward_batch_size = 512
        while Settings.backward_batch_size >= 1:
            if self.select_params(objective):
                logger.info(f'Initialized abstractor: mode="{self.mode}", method="{self.method}", input_split={self.input_split}, backward_batch_size={Settings.backward_batch_size}')
                return None
            Settings.backward_batch_size = Settings.backward_batch_size // 2

        logger.info('[!] Initialization failed')
        raise

    # @beartype
    def select_params(self: 'NetworkAbstractor', objective: typing.Any, extra_opts: dict = {}) -> bool:
        params = [
            ['patches', self.method], # default
            ['matrix', self.method],
        ]
        if self.input_split and (self.method != 'backward'):
            params += [
                ['patches', 'backward'],
                ['matrix', 'backward'],
                # ['patches', 'forward'],
                # ['matrix', 'forward'],
            ]

        for mode, method in params:
            logger.debug(f'Try conv_mode={mode}, method={method}, input_split={self.input_split}')
            self._init_module(mode=mode, objective=objective, extra_opts=extra_opts)
            if self._check_module(method=method, objective=objective):
                self.mode = mode
                self.method = method
                return True

        return False

    # @beartype
    def _init_module(self: 'NetworkAbstractor', mode: str, objective: typing.Any, extra_opts: dict = {}) -> None:
        bound_opts = {'conv_mode': mode, 'verbosity': 0, **extra_opts}
        logger.debug(f'bound_opts={bound_opts}')
        self.net = BoundedModule(
            model=self.pytorch_model,
            global_input=torch.zeros(self.input_shape, device=self.device),
            bound_opts=bound_opts,
            device=self.device,
            verbose=False,
        )
        self.net.eval()
        self.net.get_split_nodes(input_split=False)

        # check conversion correctness
        dummy = objective.lower_bounds[0].view(1, *self.input_shape[1:]).to(self.device)
        try:
            assert torch.allclose(self.pytorch_model(dummy), self.net(dummy), atol=1e-4, rtol=1e-5)
        except:
            print('[!] Conversion error')
            raise ValueError(f'torch allclose failed: {torch.norm(self.pytorch_model(dummy) - self.net(dummy))}')


    # @beartype
    def _check_module(self: 'NetworkAbstractor', method: str, objective: typing.Any) -> bool:
        # at least can run with batch=1
        x_L = objective.lower_bounds[0].view(self.input_shape)
        x_U = objective.upper_bounds[0].view(self.input_shape)
        x = self.new_input(x_L=x_L, x_U=x_U)

        # forward to save architectural information
        self.net(x)

        if math.prod(self.input_shape) >= 100000:
            return True

        try:
            self.net.set_bound_opts(get_check_abstractor_params())
            self.net.init_alpha(x=(x,)) if method == 'crown-optimized' else None
            lb, _ = self.net.compute_bounds(x=(x,), method=method) # FIXME: it uses a lot of RAM
            assert not torch.isnan(lb).any()
        except KeyboardInterrupt:
            exit()
        except SystemExit:
            exit()
        except:
            # raise
            if logger.level <= logging.DEBUG:
                traceback.print_exc()
            else:
                logger.info(f'[!] Error when trying method="{method}", backward_batch_size={Settings.backward_batch_size}')
            return False
        else:
            return True


    # @beartype
    def initialize(self: 'NetworkAbstractor', objective: typing.Any, reference_bounds: dict | None = None, init_betas: typing.Any = None) -> AbstractResults:
        objective.cs = objective.cs.to(self.device)
        objective.rhs = objective.rhs.to(self.device)

        # input property
        input_lowers = objective.lower_bounds.view(-1, *self.input_shape[1:]).to(self.device)
        input_uppers = objective.upper_bounds.view(-1, *self.input_shape[1:]).to(self.device)

        # stop function used when optimizing abstraction
        stop_criterion_func = stop_criterion_batch_any(objective.rhs)

        # create input
        x = self.new_input(x_L=input_lowers, x_U=input_uppers)

        # update initial reference bounds for later use
        self.init_reference_bounds = reference_bounds

        # get split nodes
        self.net.get_split_nodes(input_split=False)

        if self.method not in ['crown-optimized']:
            with torch.no_grad():
                lb, _ = self.net.compute_bounds(
                    x=(x,),
                    C=objective.cs,
                    method=self.method,
                    reference_bounds=reference_bounds,
                )
            logger.info(f'Initial bounds (fisrt 10): {lb.detach().cpu().flatten()[:10]}')
            if stop_criterion_func(lb).all().item():
                return AbstractResults(**{'output_lbs': lb})

            return AbstractResults(**{
                'objective_ids': getattr(objective, 'ids', None),
                'output_lbs': lb,
                'slopes': self.get_slope(),
                'lAs': self.get_lAs(),
                'cs': objective.cs,
                'rhs': objective.rhs,
                'input_lowers': input_lowers,
                'input_uppers': input_uppers,
            })

        # setup optimization parameters
        self.net.set_bound_opts(get_initialize_opt_params(stop_criterion_func))

        # initial bounds
        lb, _, aux_reference_bounds = self.net.init_alpha(
            x=(x,),
            share_alphas=Settings.share_alphas,
            c=objective.cs,
            bound_upper=False,
        )
        logger.info(f'Initial bounds (fisrt 10): {lb.detach().cpu().flatten()[:10]}')

        if stop_criterion_func(lb).all().item():
            return AbstractResults(**{'output_lbs': lb})

        # self.update_refined_beta(init_betas, batch=len(objective.cs))

        lb, _ = self.net.compute_bounds(
            x=(x,),
            C=objective.cs,
            method='crown-optimized',
            aux_reference_bounds=aux_reference_bounds,
            reference_bounds=reference_bounds,
        )
        logger.info(f'Initial optimized bounds (fisrt 10): {lb.detach().cpu().flatten()[:10]}')
        if stop_criterion_func(lb).all().item():
            return AbstractResults(**{'output_lbs': lb})

        # reorganize tensors
        with torch.no_grad():
            lower_bounds, upper_bounds = self.get_hidden_bounds(lb)

        return AbstractResults(**{
            'objective_ids': objective.ids,
            'output_lbs': lower_bounds[self.net.final_name],
            'lAs': self.get_lAs(),
            'lower_bounds': lower_bounds,
            'upper_bounds': upper_bounds,
            'slopes': self.get_slope(),
            'histories': {_.name: ([], [], []) for _ in self.net.split_nodes},
            'cs': objective.cs,
            'rhs': objective.rhs,
            'input_lowers': input_lowers,
            'input_uppers': input_uppers,
        })


    # @beartype
    def _forward_hidden(self: 'NetworkAbstractor', domain_params: AbstractResults, decisions: list, simplify: bool) -> AbstractResults:
        assert len(decisions) == len(domain_params.cs) == len(domain_params.rhs) == \
               len(domain_params.input_lowers) == len(domain_params.input_uppers), \
               f'{len(decisions)=}, {len(domain_params.input_lowers)=}'

        batch = len(decisions)
        assert batch > 0

        # 2 * batch
        double_cs = torch.cat([domain_params.cs, domain_params.cs], dim=0)
        double_input_lowers = torch.cat([domain_params.input_lowers, domain_params.input_lowers], dim=0)
        double_input_uppers = torch.cat([domain_params.input_uppers, domain_params.input_uppers], dim=0)
        if os.environ.get('NEURALSAT_ASSERT'):
            assert torch.all(double_input_lowers <= double_input_uppers)

        # update hidden bounds with new decisions (perform splitting)
        new_intermediate_layer_bounds = self.hidden_split_idx(
            lower_bounds=domain_params.lower_bounds,
            upper_bounds=domain_params.upper_bounds,
            decisions=decisions
        )

        # create new inputs
        new_x = self.new_input(x_L=double_input_lowers, x_U=double_input_uppers)

        # update slopes
        if len(domain_params.slopes) > 0:
            self.set_slope(domain_params.slopes)

        # simplify for decision heuristics
        if simplify:
            # setup optimization parameters
            self.net.set_bound_opts(get_branching_opt_params())

            # compute outputs
            with torch.no_grad():
                double_output_lbs, _, = self.net.compute_bounds(
                    x=(new_x,),
                    C=double_cs,
                    method='backward',
                    reuse_alpha=True,
                    interm_bounds=new_intermediate_layer_bounds
                )
            return AbstractResults(**{'output_lbs': double_output_lbs})

        # 2 * batch
        assert len(decisions) == len(domain_params.objective_ids)
        double_rhs = torch.cat([domain_params.rhs, domain_params.rhs], dim=0)
        double_objective_ids = torch.cat([domain_params.objective_ids, domain_params.objective_ids], dim=0)
        double_sat_solvers = domain_params.sat_solvers * 2 if domain_params.sat_solvers is not None else None

        # update new decisions
        double_histories = self.update_histories(histories=domain_params.histories, decisions=decisions)
        double_betas = domain_params.betas * 2
        num_splits = self.set_beta(betas=double_betas, histories=double_histories)

        # setup optimization parameters
        self.net.set_bound_opts(get_beta_opt_params(stop_criterion_batch_any(double_rhs)))

        # compute outputs
        double_output_lbs, _ = self.net.compute_bounds(
            x=(new_x,),
            C=double_cs,
            method=self.method,
            decision_thresh=double_rhs,
            interm_bounds=new_intermediate_layer_bounds,
        )

        # reorganize output
        with torch.no_grad():
            # lAs
            double_lAs = self.get_lAs(size=len(double_input_lowers))
            # outputs
            double_output_lbs = double_output_lbs.detach().to(device='cpu')
            # slopes
            double_slopes = self.get_slope() if len(domain_params.slopes) > 0 else {}
            # betas
            double_betas = self.get_beta(num_splits)
            # hidden bounds
            double_lower_bounds, double_upper_bounds = self.get_hidden_bounds(double_output_lbs)

        assert all([_.shape[0] == 2 * batch for _ in double_lower_bounds.values()]), print([_.shape for _ in double_lower_bounds.values()])
        assert all([_.shape[0] == 2 * batch for _ in double_upper_bounds.values()]), print([_.shape for _ in double_upper_bounds.values()])
        assert all([_.shape[0] == 2 * batch for _ in double_lAs.values()]), print([_.shape for _ in double_lAs.values()])
        assert len(double_histories) == len(double_betas) == 2 * batch

        return AbstractResults(**{
            'objective_ids': double_objective_ids,
            'output_lbs': double_lower_bounds[self.net.final_name],
            'input_lowers': double_input_lowers,
            'input_uppers': double_input_uppers,
            'lAs': double_lAs,
            'lower_bounds': double_lower_bounds,
            'upper_bounds': double_upper_bounds,
            'slopes': double_slopes,
            'betas': double_betas,
            'histories': double_histories,
            'cs': double_cs,
            'rhs': double_rhs,
            'sat_solvers': double_sat_solvers,
        })


    # @beartype
    def _forward_input(self: 'NetworkAbstractor', domain_params: AbstractResults, decisions: torch.Tensor, simplify: bool) -> AbstractResults:
        assert len(decisions) == len(domain_params.cs) == len(domain_params.rhs) == \
               len(domain_params.input_lowers) == len(domain_params.input_uppers) == len(domain_params.objective_ids)

        batch = len(decisions)
        assert batch > 0

        # splitting input by decisions (perform splitting)
        new_input_lowers, new_input_uppers = self.input_split_idx(
            input_lowers=domain_params.input_lowers,
            input_uppers=domain_params.input_uppers,
            split_idx=decisions,
        )

        # create new inputs
        new_x = self.new_input(x_L=new_input_lowers, x_U=new_input_uppers)

        # 2 * batch
        double_objective_ids = torch.cat([domain_params.objective_ids, domain_params.objective_ids], dim=0)
        double_cs = torch.cat([domain_params.cs, domain_params.cs], dim=0)
        double_rhs = torch.cat([domain_params.rhs, domain_params.rhs], dim=0)

        # set slope again since batch might change
        if len(domain_params.slopes) > 0:
            self.set_slope(domain_params.slopes)

        # set optimization parameters
        self.net.set_bound_opts(get_input_opt_params(stop_criterion_batch_any(double_rhs)))

        double_output_lbs, _ = self.net.compute_bounds(
            x=(new_x,),
            C=double_cs,
            method=self.method,
            decision_thresh=double_rhs,
            reference_bounds=self.init_reference_bounds,
        )

        with torch.no_grad():
            # slopes
            double_slopes = self.get_slope() if len(domain_params.slopes) > 0 else {}
            double_lAs = self.get_lAs(size=len(new_input_lowers))

        return AbstractResults(**{
            'objective_ids': double_objective_ids,
            'output_lbs': double_output_lbs,
            'input_lowers': new_input_lowers,
            'input_uppers': new_input_uppers,
            'slopes': double_slopes,
            'lAs': double_lAs,
            'cs': double_cs,
            'rhs': double_rhs,
        })


    # @beartype
    def forward(self: 'NetworkAbstractor', decisions: list | torch.Tensor, domain_params: AbstractResults) -> AbstractResults:
        self.iteration += 1
        forward_func = self._forward_input if self.input_split else self._forward_hidden
        return forward_func(domain_params=domain_params, decisions=decisions, simplify=False)


    def __repr__(self):
        return f'{self.__class__.__name__}({self.mode}, {self.method})'


    from .utils import (
        new_input,
        get_slope, set_slope,
        get_beta, set_beta, reset_beta, update_refined_beta,
        get_hidden_bounds,
        get_lAs,
        update_histories,
        hidden_split_idx, input_split_idx,
        build_lp_solver, solve_full_assignment,
        compute_stability,
    )



import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

import onnxruntime as ort
from pathlib import Path
import numpy as np
import onnx
import sys


from dnnv.nn.transformers.simplifiers import (simplify as dnnv_simplify, ReluifyMaxPool) # type: ignore
from dnnv.nn import parse # type: ignore
from onnxsim import simplify as onnxsim_simplify# type: ignore

def inference_onnx(path, *inputs):
    sess = ort.InferenceSession(onnx.load(path).SerializeToString())
    names = [i.name for i in sess.get_inputs()]
    inp = dict(zip(names, inputs))
    res = sess.run(None, inp)
    return res[0]

def check_conversion(input_onnx_name, output_onnx_name):
    # check conversion
    onnx_model = onnx.load(output_onnx_name)
    onnx_input_dims = onnx_model.graph.input[0].type.tensor_type.shape.dim
    onnx_input_shape = tuple(d.dim_value for d in onnx_input_dims)
    dummy = np.random.randn(*onnx_input_shape).astype(np.float32)
    o1 = inference_onnx(input_onnx_name, dummy)
    o2 = inference_onnx(output_onnx_name, dummy)
    # print(np.sum(np.abs(o1 - o2)))
    return np.allclose(o1, o2, rtol=1e-4, atol=1e-5)


def dnnv_convert(input_onnx_name, output_onnx_name):
    os.system(f'rm -rf {output_onnx_name}')

    # load  model
    op_graph = parse(Path(input_onnx_name))

    # convert model
    sim_op_graph = dnnv_simplify(op_graph, ReluifyMaxPool(op_graph))

    # export model
    sim_op_graph.export_onnx(output_onnx_name)

    # check
    if check_conversion(input_onnx_name, output_onnx_name):
        print("[+] DNNV exported to:", output_onnx_name)
        return True

    print(f"[!] DNNV failed to simplify {input_onnx_name}.")
    os.system(f'rm -rf {output_onnx_name}')
    return False


def onnxsim_convert(input_onnx_name, output_onnx_name):
    os.system(f'rm -rf {output_onnx_name}')

    # load  model
    model = onnx.load(input_onnx_name)

    # convert model
    model_simp, check = onnxsim_simplify(model)

    # check
    if not check:
        print(f"[!] ONNXSIM failed to simplify {input_onnx_name}.")
        os.system(f'rm -rf {output_onnx_name}')
        return False

    # export model
    onnx.save(model_simp, output_onnx_name)

    # check
    if check_conversion(input_onnx_name, output_onnx_name):
        print("[+] ONNXSIM exported to:", output_onnx_name)
        return True

    print(f"[!] ONNXSIM failed to simplify {input_onnx_name}.")
    os.system(f'rm -rf {output_onnx_name}')
    return False


def convert(input_onnx_name, output_onnx_name_wo_ext):
    os.makedirs(os.path.dirname(output_onnx_name_wo_ext), exist_ok=True)

    dnnv_output = output_onnx_name_wo_ext + '-dnnv.onnx'
    onnxsim_output = output_onnx_name_wo_ext + '-onnxsim.onnx'

    check_dnnv = dnnv_convert(input_onnx_name, dnnv_output)
    if check_dnnv:
        check_onnxsim = onnxsim_convert(dnnv_output, onnxsim_output)
    else:
        check_onnxsim = onnxsim_convert(input_onnx_name, onnxsim_output)

    return (check_dnnv or check_onnxsim)


if __name__ == "__main__":
    assert len(sys.argv) == 3, "expected 1 arguments: [input w/ .onnx, output w/o .onnx]"
    convert(sys.argv[1], sys.argv[2])
    # python3 util/network/simplify_onnx.py example/onnx/mnist-net_256x2.onnx ./output/mnist-simplified



from itertools import accumulate

import torch

from ..operations.base import Operator
from ..utils import assign_values_to_dim


class Split(Operator):
    def __init__(
        self,
        enable_pruning=False,
        split_size_or_sections=None,
        number_of_splits=None,
        dim=0,
        keep_size=True,
    ):
        """
        Parameters
        ----------
        enable_pruning: bool
        split_size_or_sections: tuple[int]
        number_of_splits: int
            The number of equal splits along dim.
        dim: int
            Split dimension. Tensor is split over this axis.
        keep_size: bool
            If True it keeps the size of the split the same as in initial pass.
            Else it splits it accordingly to the pruned input.
        """
        if enable_pruning:
            assert (
                split_size_or_sections is not None or number_of_splits is not None
            ), "One of the parameters needs to be set."
        self.enable_pruning = enable_pruning
        self.dim = dim
        self.split_size_or_sections = split_size_or_sections
        # self.split_size_or_sections = (split_size_or_sections[0], 1)
        self.number_of_splits = number_of_splits
        self.keep_size = keep_size
        self.input_indices = None
        self.placeholder = None
        super().__init__()

    def _get_sections(self, input):
        """Calculate sections from number of splits."""
        dim_size = input[0].shape[self.dim]
        assert (
            dim_size % self.number_of_splits == 0
        ), "Dimension size {} not equally divisible by {}.".format(
            dim_size, self.number_of_splits
        )
        s = dim_size // self.number_of_splits
        sections = tuple(s for _ in range(self.number_of_splits))
        return sections

    def forward(self, *input):
        if not self.enable_pruning and len(input) == 2:
            return torch.split(input[0], list(input[1]), dim=self.dim)
        if self.split_size_or_sections is None:
            self.split_size_or_sections = self._get_sections(input)

        if self.input_indices is not None:
            self.placeholder *= 0
            assign_values_to_dim(
                self.placeholder, input[0], self.input_indices, self.dim
            )
            split = torch.split(self.placeholder, self.split_size_or_sections, self.dim)
        else:
            split = torch.split(*input, self.split_size_or_sections, dim=self.dim)
        return split

    def set_input_indices(self, input: tuple):
        assert isinstance(input, (tuple, list))

        inp = input[0]
        # We assume that aggregation dimensions correspond to split dimension
        axis = self.get_axis(inp.shape, self.dim)

        # Mask shows where features are non zero in the whole axis.
        mask = inp != 0
        if len(inp.shape) > 1:
            mask = mask.sum(axis=tuple(axis)) != 0

        if not self.keep_size:
            # Read docstrings
            if isinstance(self.split_size_or_sections, tuple):
                indices = list(accumulate(self.split_size_or_sections))
                indices = torch.tensor(indices) - 1
            else:
                raise NotImplementedError("Not implemented for split size.")
            cs = torch.cumsum(mask, 0)
            ind = [0] + cs[indices].tolist()
            sec = [ind[i + 1] - ind[i] for i in range(len(ind) - 1)]
            self.split_size_or_sections = sec
        else:
            (self.input_indices,) = torch.where(mask)
            self.placeholder = torch.zeros(inp.shape)

    def __str__(self):
        return "Split(dim={})".format(self.dim)



import torch
from torch import nn


class Where(nn.Module):
    def forward(self, condition: torch.Tensor, X: torch.Tensor, Y=torch.Tensor):
        res_type = torch.result_type(X, Y)
        output = torch.where(condition, X.to(res_type), Y.to(res_type))
        return output



import warnings
import os
import torch
from torch.onnx.utils import _optimize_graph
# from torch.onnx.symbolic_helper import _set_opset_version
from torch.onnx._globals import GLOBALS
from collections import OrderedDict
from collections import namedtuple
import re
from .bounded_tensor import BoundedTensor, BoundedParameter
from .utils import logger, unpack_inputs

Node = namedtuple('Node', (
    'name', 'ori_name', 'inputs', 'attr', 'op', 'param', 'input_index',
    'bound_node', 'output_index', 'perturbation'), defaults=(None,) * 10)

def get_node_name(node):
    return node.debugName()

def get_node_attribute(node, attribute_name):
    if hasattr(torch.onnx.symbolic_helper, '_node_get'):
        # Pytorch >= 1.13.
        return torch.onnx.symbolic_helper._node_get(node, attribute_name)
    else:
        # Pytorch <= 1.12. This will call _node_getitem in torch.onnx.utils.
        return node[attribute_name]

def parse_graph(graph, inputs, params):
    input_all = []
    input_used = []
    scope = {}
    for n in graph.inputs():
        input_all.append(n.debugName())
    for n in graph.nodes():
        n_inputs = [get_node_name(i) for i in n.inputs()]
        for inp in n.inputs():
            input_used.append(inp.debugName())
        for out in n.outputs():
            scope[get_node_name(out)] = n.scopeName()
    for node in graph.inputs():
        name = get_node_name(node)
        scope[name] = ''
    for n in graph.outputs():
        name = get_node_name(n)
        if name in input_all:
            # This output node directly comes from an input node with an Op
            input_used.append(n.debugName())

    def name_with_scope(node):
        name = get_node_name(node)
        return '/'.join([scope[name], name])

    nodesOP = []
    for n in graph.nodes():
        attrs = {k: get_node_attribute(n, k) for k in n.attributeNames()}
        n_inputs = [name_with_scope(i) for i in n.inputs()]
        for i, out in enumerate(list(n.outputs())):
            nodesOP.append(Node(**{'name': name_with_scope(out),
                                'op': n.kind(),
                                'inputs': n_inputs,
                                'attr': attrs,
                                'output_index': i,
                                }))

    # filter out input nodes in `graph.inputs()` that are actually used
    nodesIn = []
    used_by_index = []
    for i, n in enumerate(graph.inputs()):
        name = get_node_name(n)
        used = name in input_used
        used_by_index.append(used)
        if used:
            nodesIn.append(n)
    # filter out input nodes in `inputs` that are actually used
    inputs_unpacked = unpack_inputs(inputs)
    assert len(list(graph.inputs())) == len(inputs_unpacked) + len(params)
    inputs = [inputs_unpacked[i] for i in range(len(inputs_unpacked)) if used_by_index[i]]
    # index of the used inputs among all the inputs
    input_index = [i for i in range(len(inputs_unpacked)) if used_by_index[i]]
    # Add a name to all inputs
    inputs = list(zip(["input_{}".format(input_index[i]) for i in range(len(inputs))], inputs))
    # filter out params that are actually used
    params = [params[i] for i in range(len(params)) if used_by_index[i + len(inputs_unpacked)]]
    inputs_and_params = inputs + params
    assert len(nodesIn) == len(inputs_and_params)

    # output nodes of the module
    nodesOut = []
    for n in graph.outputs():
        # we only record names
        nodesOut.append(name_with_scope(n))

    for i, n in enumerate(nodesIn):
        if (isinstance(inputs_and_params[i][1], BoundedTensor) or
                isinstance(inputs_and_params[i][1], BoundedParameter)):
            perturbation = inputs_and_params[i][1].ptb
        else:
            perturbation = None
        if i > 0 and n.type().sizes() != list(inputs_and_params[i][1].size()):
            raise RuntimeError("Input tensor shapes do not much: {} != {}".format(
                n.type().sizes(), list(inputs_and_params[i][1].size())))
        nodesIn[i] = Node(**{'name': name_with_scope(n),
                             'ori_name': inputs_and_params[i][0],
                             'op': 'Parameter',
                             'inputs': [],
                             'attr': str(n.type()),
                             'param': inputs_and_params[i][1] if i >= len(inputs) else None,
                             # index among all the inputs including unused ones
                             'input_index': input_index[i] if i < len(inputs) else None,
                             # Input nodes may have perturbation, if they are wrapped in BoundedTensor or BoundedParameters
                             'perturbation': perturbation, })

    return nodesOP, nodesIn, nodesOut

def _get_jit_params(module, param_exclude, param_include):
    state_dict = torch.jit._unique_state_dict(module, keep_vars=True)

    if param_exclude is not None:
        param_exclude = re.compile(param_exclude)
    if param_include is not None:
        param_include = re.compile(param_include)

    new_state_dict = OrderedDict()
    for k, v in state_dict.items():
        if param_exclude is not None and param_exclude.match(k) is not None:
            print(f'\nremove input element {k} from nodesIn\n')
            continue
        if param_include is not None and param_include.match(k) is None:
            continue
        new_state_dict[k] = v

    params = zip(new_state_dict.keys(), new_state_dict.values())

    return params

def get_output_template(out):
    """Construct a template for the module output with `None` representing places
    to be filled with tensor results"""
    if isinstance(out, torch.Tensor):
        return None
    elif isinstance(out, list):
        return list([get_output_template(o) for o in out])
    elif isinstance(out, tuple):
        return tuple([get_output_template(o) for o in out])
    elif isinstance(out, dict):
        template = {}
        for key in out:
            template[key] = get_output_template(out[key])
        return template
    else:
        raise NotImplementedError

def parse_module(module, inputs, param_exclude=".*AuxLogits.*", param_include=None):
    params = _get_jit_params(module, param_exclude=param_exclude, param_include=param_include)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        trace, out = torch.jit._get_trace_graph(module, inputs)
    
    # _set_opset_version(12) # FIXME: attrs might differ 
    GLOBALS._export_onnx_opset_version = 12
    
    # Assuming that the first node in the graph is the primary input node.
    # It must have a batch dimension.
    primary_input = get_node_name(next(iter(trace.inputs())))
    trace_graph = _optimize_graph(
        trace, torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK,
        params_dict={},
        input_names=[primary_input],
        dynamic_axes={primary_input: {0: 'batch'}})
    logger.debug('trace_graph: %s', trace_graph)

    if int(os.environ.get('AUTOLIRPA_DEBUG_GRAPH', 0)) > 0:
        print("Graph before ONNX convertion:")
        print(trace)
        print("ONNX graph:")
        print(trace_graph)

    if not isinstance(inputs, tuple):
        inputs = (inputs, )

    nodesOP, nodesIn, nodesOut = parse_graph(trace_graph, tuple(inputs), tuple(params))

    for i in range(len(nodesOP)):
        param_in = OrderedDict()
        for inp in nodesOP[i].inputs:
            for n in nodesIn:
                if inp == n.name:
                    param_in.update({inp:n.param})
        nodesOP[i] = nodesOP[i]._replace(param=param_in)

    template = get_output_template(out)

    return nodesOP, nodesIn, nodesOut, template



from pathlib import Path
import numpy as np
import time
import os

import warnings
warnings.filterwarnings(action='ignore')

from verifier.objective import Objective, DnfObjectives
from util.spec.read_vnnlib import read_vnnlib
from util.network.read_onnx import parse_onnx
from verifier.verifier import Verifier 
from util.misc.result import ReturnStatus
from util.misc.logger import logger
from auto_LiRPA.bound_ops import BoundRelu

def extract_instance(net_path, vnnlib_path):
    vnnlibs = read_vnnlib(vnnlib_path)
    model, input_shape, output_shape, is_nhwc = parse_onnx(net_path)
    
    # objective
    objectives = []
    for spec in vnnlibs:
        bounds = spec[0]
        for prop_i in spec[1]:
            objectives.append(Objective((bounds, prop_i)))
    objectives = DnfObjectives(objectives, input_shape=input_shape, is_nhwc=is_nhwc)

    return model, input_shape, objectives


def refine_layer(node):
    from auto_LiRPA.bound_ops import BoundLinear
    
    for n in node.inputs:
        refine_layer(n)
        
    if isinstance(node, BoundLinear):
        print('[+] Refine layer:', node)
        # print('\t- lower:', node.lower)
        # print('\t- upper:', node.upper)
        candidates = []
        candidate_neuron_ids = []
        
        for neuron_idx, v in enumerate(node.solver_vars):
            # print(v.VarName, v.lb==node.lower[0, neuron_idx], v.ub==node.upper[0, neuron_idx] if node.upper is not None else None)
            candidates.append(v.VarName)
            candidate_neuron_ids.append(neuron_idx)
            v.lb = -np.inf
            v.ub = np.inf
        
        # exit()
    

if __name__ == "__main__":
    
    net_path = 'example/test_mnistfc.onnx'
    # net_path = 'example/mnistfc-medium-net-151.onnx'
    vnnlib_path = Path('example/prop_2_0.03.vnnlib')
    
    net_path = '../benchmark/mnistfc_hard/onnx/mnistfc-hard-net-177.onnx'
    vnnlib_path = Path('../benchmark/mnistfc_hard/spec/prop_8_0.05.vnnlib')
    
    print('\nRunning test with', net_path, vnnlib_path)
    device = 'cpu'
    batch = 1
    logger.setLevel(1)
    
    preconditions = [eval(line) for line in open('clause.txt').read().strip().split('\n') if not line.startswith('#')]
    # preconditions = []
    print(preconditions)
    

    model, input_shape, objectives = extract_instance(net_path, vnnlib_path)
    model.to(device)
    print(model)
    
    verifier = Verifier(
        net=model, 
        input_shape=input_shape, 
        batch=batch,
        device=device,
    )

    
    print(verifier.verify(objectives, preconditions=preconditions))
    exit()
    
    # obj = objectives.pop(1)
    # input_lowers = obj.lower_bounds.view(input_shape).to(device)
    # input_uppers = obj.upper_bounds.view(input_shape).to(device)
    # c = obj.cs.to(device)
    # print(input_uppers.shape)
    # print(obj.cs.shape)
    
    
    # verifier._setup_restart(0, obj)
    # verifier.abstractor.initialize(obj, None)
    
    # tic = time.time()
    # verifier.abstractor.build_lp_solver('mip', input_lowers, input_uppers, c=None)
    # print('MIP refine:', time.time() - tic)
    
    # intermediate_layer_bounds = verifier.abstractor.net.get_refined_interm_bounds()
    
    # for k, v in intermediate_layer_bounds.items():
    #     print(k, [_.shape for _ in v])
    # exit()
    
    # verifier.abstractor.initialize(obj, None, intermediate_layer_bounds)
    
    # # name_dict = {i: layer.inputs[0].name for (i, layer) in enumerate(verifier.abstractor.net.perturbed_optimizable_activations)}
    # # pre_relu_indices = [i for (i, layer) in enumerate(verifier.abstractor.net.perturbed_optimizable_activations) if isinstance(layer, BoundRelu)]
    # # print(pre_relu_indices)
    # # print(name_dict)
    # for _ in range(0):
    #     tic = time.time()
    #     intermediate_layer_bounds = verifier.abstractor.net.get_refined_interm_bounds()
    #     verifier.abstractor.build_lp_solver('mip', input_lowers, input_uppers, c=None, intermediate_layer_bounds=intermediate_layer_bounds)
    #     print(_, 'MIP refine:', time.time() - tic)
        
    #     intermediate_layer_bounds = verifier.abstractor.net.get_refined_interm_bounds()
    #     verifier.abstractor.initialize(obj, None, intermediate_layer_bounds)
    #     print()
    
    
    # tic = time.time()
    # intermediate_layer_bounds[name_dict[pre_relu_indices[1]]][0][0][83] = -0.02
    # intermediate_layer_bounds[name_dict[pre_relu_indices[2]]][1][0][26] = -0.05
    # intermediate_layer_bounds[name_dict[4]][1][0][105] = 2.05
    # verifier.abstractor.build_lp_solver('mip', input_lowers, input_uppers, c=None, intermediate_layer_bounds=intermediate_layer_bounds)
    # print('MIP refine:', time.time() - tic)
    
    # verifier.abstractor.net._reset_solver_vars(verifier.abstractor.net.final_node())
    # del verifier.abstractor.net.model
    
    # verifier.abstractor.build_lp_solver('lp', input_lowers, input_uppers, c)
    # print(verifier.abstractor.net.model)
    
    
    # for v in verifier.abstractor.net.model.getVars():
    #     print(v.VarName, v.lb, v.ub)
    
    # refine_layer(verifier.abstractor.net.final_node())
    # verifier.abstractor.net.model.update()
    
    # for v in verifier.abstractor.net.model.getVars():
    #     print(v.VarName, v.lb, v.ub)
        


from beartype import beartype
import random
import torch
import os

from ..onnx2pytorch.convert.model import ConvertModel
from ..verifier.objective import DnfObjectives

from ..util.misc.torch_cuda_memory import is_cuda_out_of_memory, gc_cuda
from ..util.misc.check import check_solution
from ..util.misc.logger import logger

from .random_attack import RandomAttacker
from .pgd_attack.general import attack


import pdb
DBG = pdb.set_trace

class Attacker:

    # @beartype
    def __init__(self: 'Attacker', net: ConvertModel, objective: DnfObjectives, input_shape: tuple, device: str) -> None:
        self.attackers = [
            RandomAttacker(net, objective, input_shape, device=device),
            PGDAttacker(net, objective, input_shape, device=device),
        ]

    # @beartype
    def run(self: 'Attacker', timeout: float = 0.5) -> tuple[bool, torch.Tensor | None]:
        return self._attack(timeout=timeout)

    # @beartype
    def _attack(self: 'Attacker', timeout: float) -> tuple[bool, torch.Tensor | None]:
        for atk in self.attackers:
            seed = random.randint(0, 1000)
            atk.manual_seed(seed)
            try:
                # attacker using float64 might get OOM
                is_attacked, adv = atk.run(timeout=timeout)
            except RuntimeError as exception:
                if is_cuda_out_of_memory(exception):
                    # restore to default data type
                    atk.net.to(torch.get_default_dtype())
                    logger.info(f"[Failed] {atk} got OOM")
                    return False, None
                else:
                    raise NotImplementedError
            except:
                raise NotImplementedError
            else:
                gc_cuda()
            logger.info(f"{'[Success]' if is_attacked else '[Failed]'} {atk}")
            if is_attacked:
                return is_attacked, adv
        return False, None


class PGDAttacker:

    # @beartype
    def __init__(self: 'PGDAttacker', net: ConvertModel, objective: DnfObjectives, input_shape: tuple, device: str = 'cpu') -> None:
        self.net = net
        self.objective = objective
        self.input_shape = input_shape
        self.device = device
        self.seed = None

    # @beartype
    def manual_seed(self: 'PGDAttacker', seed: int) -> None:
        self.seed = seed
        random.seed(self.seed)
        torch.manual_seed(self.seed)


    # @beartype
    def run(self: 'PGDAttacker', iterations: int = 50, restarts: int = 20, timeout: float = 2.0) -> tuple[bool, torch.Tensor | None]:
        data_min = self.objective.lower_bounds.view(-1, *self.input_shape[1:]).unsqueeze(0).to(self.device)
        data_max = self.objective.upper_bounds.view(-1, *self.input_shape[1:]).unsqueeze(0).to(self.device)

        data_min_f64 = self.objective.lower_bounds_f64.view(-1, *self.input_shape[1:]).unsqueeze(0).to(self.device)
        data_max_f64 = self.objective.upper_bounds_f64.view(-1, *self.input_shape[1:]).unsqueeze(0).to(self.device)

        # assert torch.all(data_min <= data_max)
        # x = (data_min[:, 0] + data_max[:, 0]) / 2
        x = (data_max[:, 0] - data_min[:, 0]) * torch.rand(data_min[:, 0].shape, device=self.device) + data_min[:, 0]
        if os.environ.get('NEURALSAT_ASSERT'):
            assert torch.all(x <= data_max[:, 0])
            assert torch.all(x >= data_min[:, 0])

        cs = self.objective.cs.to(self.device)
        rhs = self.objective.rhs.to(self.device)

        cs_f64 = self.objective.cs_f64.to(self.device)
        rhs_f64 = self.objective.rhs_f64.to(self.device)

        try:
            self.net.to(cs_f64.dtype)
            is_attacked, attack_images = attack(
                model=self.net,
                x=x.to(cs_f64.dtype),
                data_min=data_min_f64,
                data_max=data_max_f64,
                cs=cs_f64,
                rhs=rhs_f64,
                attack_iters=iterations,
                num_restarts=restarts,
                timeout=timeout,
            )
        except RuntimeError as exception:
            if is_cuda_out_of_memory(exception):
                self.net.to(cs.dtype)
                is_attacked, attack_images = attack(
                    model=self.net,
                    x=x.to(cs.dtype),
                    data_min=data_min,
                    data_max=data_max,
                    cs=cs,
                    rhs=rhs,
                    attack_iters=iterations,
                    num_restarts=restarts,
                    timeout=timeout,
                )
            else:
                raise NotImplementedError()
        except:
            raise NotImplementedError()


        if is_attacked:
            with torch.no_grad():
                for i in range(attack_images.shape[1]): # restarts
                    for j in range(attack_images.shape[2]): # props
                        adv = attack_images[:, i, j]
                        if check_solution(self.net, adv, cs=cs_f64[j], rhs=rhs_f64[j], data_min=data_min_f64[:, j], data_max=data_max_f64[:, j]):
                            return True, adv
            logger.debug("[!] Invalid counter-example")

        self.net.to(cs.dtype)
        return False, None


    def __str__(self):
        return f'PGDAttack(seed={self.seed}, device={self.device})'
